# basic imports
import numpy as np

# 3rd party imports
from PyQt6.QtWidgets import *
from PyQt6.QtCore import *
from PyQt6.QtGui import *

# self imports
from grid.grid import *
from grid.gui.customQt import *


class PnKmeaner(QWidget):
    """ """

    def __init__(self, grid):
        super().__init__()
        self.setFocusPolicy(Qt.StrongFocus)
        self.setFocus()
        self.update()
        """attr"""
        self.grid = grid
        # # grid params.
        self.nFeatures = min(self.grid.imgs.depth, 9)
        self.features = []
        self.lsSelect = None

        # main/img
        self.layout = QHBoxLayout()
        # img preview (left)
        self.gr_left = QGroupBox()
        self.lo_left = QGridLayout()
        self.wg_img = Widget_Kmeans(grid)
        self.bt_ccw = QPushButton("rotate ccw (Q)")
        self.bt_cw = QPushButton("rorate cw (E)")
        # K mean (right)
        self.kMax = 9
        self.gr_pre = QGroupBox("K-means Algo.")
        self.lo_pre = QGridLayout()

        self.gr_ft = QGroupBox("Channels used for clustering")
        self.lo_ft = QHBoxLayout()
        self.ck_ft = []
        for i in range(self.nFeatures):
            checkbox = QCheckBox(str(i + 1))
            # if i < 3:
            checkbox.setChecked(True)
            checkbox.stateChanged.connect(self.change_k)
            self.ck_ft.extend([checkbox])
            self.lo_ft.addWidget(self.ck_ft[i])

        self.lsSelect = [0]
        self.gr_k = QGroupBox("K = 3")
        self.lo_k = QVBoxLayout()
        self.sl_k = QSlider(Qt.Horizontal)
        # Binarization
        self.gr_bin = QGroupBox("Binarization")
        self.lo_bin = QVBoxLayout()
        # Binarization (auto)
        self.gr_cut = QGroupBox("Auto cutoff = 1")
        self.lo_cut = QVBoxLayout()
        self.sl_cut = QSlider(Qt.Horizontal)
        self.gr_cusb = QGroupBox("Custom")
        self.lo_cusb = QHBoxLayout()
        self.ck_cusb = []
        for i in range(1, self.kMax + 1):
            checkbox = QCheckBox(str(i))
            checkbox.stateChanged.connect(self.custom_cut)
            if i > 3:
                checkbox.setEnabled(False)
            self.ck_cusb.extend([checkbox])
        self.ls_bin = [0]
        # Display
        self.gr_dis = QGroupBox("Display")
        self.lo_dis = QHBoxLayout()
        self.rb_bin = QRadioButton("Binary (A)")
        self.rb_rgb = QRadioButton("RGB (S)")
        self.rb_k = QRadioButton("K-Means (D)")
        # zoom
        self.gr_zm = QGroupBox("Magnification Levels (Right-click to switch)")
        self.lo_zm = QHBoxLayout()
        self.rb_1x = QRadioButton("1X")
        self.rb_15x = QRadioButton("1.5X")
        self.rb_3x = QRadioButton("3X")

        # refine (right)
        self.gr_pro = QGroupBox("Clusters Refine")
        self.lo_pro = QVBoxLayout()
        self.gr_shad = QGroupBox("De-Shade = 0")
        self.lo_shad = QVBoxLayout()
        self.sl_shad = QSlider(Qt.Horizontal)
        self.gr_gb = QGroupBox("De-Noise = 0")
        self.lo_gb = QVBoxLayout()
        self.sl_gb = QSlider(Qt.Horizontal)

        # panel right
        self.pn_right = QWidget()
        self.lo_right = QVBoxLayout()
        self.sc_right = QScrollArea()
        self.sc_right.setStyleSheet("QScrollBar {width:0px;}")
        self.sc_right.setWidgetResizable(True)

        """initialize UI"""
        self.initUI()

    def initUI(self):
        """img preview (left)"""
        # components
        self.bt_ccw.clicked.connect(self.rotateCCW)
        self.bt_cw.clicked.connect(self.rotateCW)
        # layout
        self.lo_left.addWidget(self.wg_img, 0, 0, 1, 2)
        self.lo_left.addWidget(self.bt_ccw, 1, 0)
        self.lo_left.addWidget(self.bt_cw, 1, 1)
        self.gr_left.setLayout(self.lo_left)
        """pre keans (right)"""
        # components
        self.sl_k.setMinimum(2)
        self.sl_k.setMaximum(self.kMax)
        self.sl_k.setValue(3)
        self.sl_k.setTickInterval(1)
        self.sl_k.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.sl_k.valueChanged.connect(self.change_k)

        # layout
        self.gr_ft.setLayout(self.lo_ft)
        self.lo_k.addWidget(self.sl_k)
        self.gr_k.setLayout(self.lo_k)
        self.lo_pre.addWidget(self.gr_ft)
        self.lo_pre.addWidget(self.gr_k)
        self.gr_pre.setLayout(self.lo_pre)

        """binarization"""
        # component
        self.gr_cut.setCheckable(True)
        self.gr_cut.setChecked(True)
        self.gr_cut.clicked.connect(self.auto_cut)
        self.sl_cut.setMinimum(1)
        self.sl_cut.setMaximum(3)
        self.sl_cut.setValue(1)
        self.sl_cut.setTickInterval(1)
        self.sl_cut.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.sl_cut.valueChanged.connect(self.auto_cut)
        self.gr_cusb.setCheckable(True)
        self.gr_cusb.setChecked(False)
        self.gr_cusb.clicked.connect(self.custom_cut)
        # layout
        self.lo_cut.addWidget(self.sl_cut)
        self.gr_cut.setLayout(self.lo_cut)
        for i in range(self.kMax):
            self.lo_cusb.addWidget(self.ck_cusb[i])
        self.gr_cusb.setLayout(self.lo_cusb)
        self.lo_bin.addWidget(self.gr_cut)
        self.lo_bin.addWidget(self.gr_cusb)
        self.gr_bin.setLayout(self.lo_bin)

        """pro keans (right)"""
        # components
        self.sl_shad.setMinimum(0)
        self.sl_shad.setMaximum(255)
        self.sl_shad.setValue(0)
        self.sl_shad.setTickInterval(20)
        self.sl_shad.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.sl_shad.valueChanged.connect(self.change_shad)
        self.sl_gb.setMinimum(0)
        self.sl_gb.setMaximum(50)
        self.sl_gb.setValue(0)
        self.sl_gb.setTickInterval(5)
        self.sl_gb.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.sl_gb.valueChanged.connect(self.change_gb)
        # layout
        self.lo_shad.addWidget(self.sl_shad)
        self.gr_shad.setLayout(self.lo_shad)
        self.lo_gb.addWidget(self.sl_gb)
        self.gr_gb.setLayout(self.lo_gb)
        self.lo_pro.addWidget(self.gr_shad)
        self.lo_pro.addWidget(self.gr_gb)
        self.gr_pro.setLayout(self.lo_pro)

        "Zoom"
        self.rb_1x.toggled.connect(lambda: self.changeZoom(0))
        self.rb_15x.toggled.connect(lambda: self.changeZoom(1))
        self.rb_3x.toggled.connect(lambda: self.changeZoom(2))
        self.lo_zm.addWidget(self.rb_1x)
        self.lo_zm.addWidget(self.rb_15x)
        self.lo_zm.addWidget(self.rb_3x)
        self.gr_zm.setLayout(self.lo_zm)

        """display"""
        # components
        self.rb_bin.setChecked(True)
        self.rb_bin.toggled.connect(self.wg_img.switch_imgB)
        self.rb_rgb.toggled.connect(self.wg_img.switch_imgVis)
        self.rb_k.toggled.connect(self.wg_img.switch_imgK)
        # layout
        self.lo_dis.addWidget(self.rb_bin)
        self.lo_dis.addWidget(self.rb_rgb)
        self.lo_dis.addWidget(self.rb_k)
        self.gr_dis.setLayout(self.lo_dis)

        """right"""
        self.lo_right.addWidget(self.gr_pre)
        self.lo_right.addWidget(self.gr_bin)
        self.lo_right.addWidget(self.gr_pro)
        self.lo_right.addWidget(self.gr_dis)
        self.lo_right.addWidget(self.gr_zm)
        self.pn_right.setLayout(self.lo_right)
        self.sc_right.setWidget(self.pn_right)

        """assemble"""
        policy_right = QSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )
        policy_right.setHorizontalStretch(2)
        self.sc_right.setSizePolicy(policy_right)
        policy_left = QSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )
        policy_left.setHorizontalStretch(3)
        self.gr_left.setSizePolicy(policy_left)
        self.layout.addWidget(self.gr_left)
        self.layout.addWidget(self.sc_right)
        self.setLayout(self.layout)
        self.change_k()  # initialize kmeans
        self.show()

        """collapse"""

    #     self.gr_pre.setCheckable(True)
    #     self.gr_pre.setChecked(True)
    #     self.gr_pre.toggled.connect(self.collapsePre)
    #     # self.collapsePre()

    #     self.gr_bin.setCheckable(True)
    #     self.gr_bin.setChecked(False)
    #     self.gr_bin.toggled.connect(self.collapseBin)
    #     self.collapseBin()

    #     self.gr_pro.setCheckable(True)
    #     self.gr_pro.setChecked(False)
    #     self.gr_pro.toggled.connect(self.collapsePro)
    #     self.collapsePro()

    #     self.gr_dis.setCheckable(True)
    #     self.gr_dis.setChecked(True)
    #     self.gr_dis.toggled.connect(self.collapseDis)
    #     # self.collapseDis()

    #     self.gr_zm.setCheckable(True)
    #     self.gr_zm.setChecked(False)
    #     self.gr_zm.toggled.connect(self.collapseZm)
    #     self.collapseZm()

    # def collapsePre(self):
    #     self.gr_ft.setVisible(not self.gr_ft.isVisible())
    #     self.gr_k.setVisible(not self.gr_k.isVisible())

    # def collapseBin(self):
    #     self.gr_cut.setVisible(not self.gr_cut.isVisible())
    #     self.gr_cusb.setVisible(not self.gr_cusb.isVisible())

    # def collapsePro(self):
    #     self.gr_shad.setVisible(not self.gr_shad.isVisible())
    #     self.gr_gb.setVisible(not self.gr_gb.isVisible())

    # def collapseDis(self):
    #     self.rb_bin.setVisible(not self.rb_bin.isVisible())
    #     self.rb_rgb.setVisible(not self.rb_rgb.isVisible())
    #     self.rb_k.setVisible(not self.rb_k.isVisible())

    # def collapseZm(self):
    #     self.rb_1x.setVisible(not self.rb_1x.isVisible())
    #     self.rb_15x.setVisible(not self.rb_15x.isVisible())
    #     self.rb_3x.setVisible(not self.rb_3x.isVisible())

    def changeZoom(self, index):
        self.wg_img.zoom = index

    def rotateCCW(self):
        self.grid.rotateImg(nRot=1)
        self.refresh()

    def rotateCW(self):
        self.grid.rotateImg(nRot=3)
        self.refresh()

    def binarizeImgGUI(self):
        self.grid.binarizeImg(
            k=self.sl_k.value(),
            features=self.features,
            lsSelect=self.lsSelect,
            valShad=self.sl_shad.value(),
            valSmth=self.sl_gb.value(),
        )
        self.refresh()

    def change_k(self):
        value = self.sl_k.value()
        ls_ft = []
        for i in range(len(self.ck_ft)):
            # if i < self.nFeatures:
            #     self.ck_ft[i].setVisible(True)
            # else:
            #     self.ck_ft[i].setVisible(False)
            if i in range(self.nFeatures) and self.ck_ft[i].isChecked():
                ls_ft.extend([i])

        self.features = ls_ft
        self.sl_cut.setMaximum(value)
        self.gr_k.setTitle("K = %d" % value)
        # will certainly go to either of cut function, no need to do binarize
        if self.gr_cusb.isChecked():
            self.custom_cut()
        else:
            self.auto_cut()

    def auto_cut(self):
        self.gr_cut.setChecked(True)
        self.gr_cusb.setChecked(False)
        value = self.sl_cut.value()
        self.gr_cut.setTitle("Auto cutoff = %d" % value)
        ls_bin = []
        for i in range(self.kMax):
            self.ck_cusb[i].setEnabled(False)
            if i < value:
                ls_bin.extend([i])
        self.lsSelect = ls_bin
        self.binarizeImgGUI()

    def custom_cut(self):
        self.gr_cut.setChecked(False)
        self.gr_cusb.setChecked(True)
        value = self.sl_k.value()
        ls_bin = []
        for i in range(self.kMax):
            if i < value:
                self.ck_cusb[i].setEnabled(True)
                if self.ck_cusb[i].isChecked():
                    ls_bin.extend([i])
            else:
                self.ck_cusb[i].setEnabled(False)
        self.lsSelect = ls_bin
        self.binarizeImgGUI()

    def change_shad(self):
        value = self.sl_shad.value()
        self.gr_shad.setTitle("De-Shade = %d" % value)
        self.binarizeImgGUI()

    def change_gb(self):
        value = self.sl_gb.value()
        self.gr_gb.setTitle("De-Noise = %d" % value)
        self.binarizeImgGUI()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_A:
            self.rb_bin.setChecked(True)
        elif event.key() == Qt.Key_S:
            self.rb_rgb.setChecked(True)
        elif event.key() == Qt.Key_D:
            self.rb_k.setChecked(True)
        elif event.key() == Qt.Key_Q:
            self.rotateCCW()
        elif event.key() == Qt.Key_E:
            self.rotateCW()

    # def keyReleaseEvent(self, event):
    # self.rb_bin.setChecked(True)

    def refresh(self):
        self.rb_bin.setChecked(True)
        self.wg_img.switch_imgB()
        # elif self.rb_rgb.isChecked():
        #     self.wg_img.switch_imgVis()
        # elif self.rb_k.isChecked():
        #     self.wg_img.switch_imgK()

    def run(self):
        # GRID got everything done
        return 0

    def paintEvent(self, paint_event):
        if self.wg_img.zoom == 0:
            self.rb_1x.setChecked(True)
        elif self.wg_img.zoom == 1:
            self.rb_15x.setChecked(True)
        else:
            self.rb_3x.setChecked(True)


class Widget_Kmeans(Widget_Img):
    def __init__(self, grid):
        super().__init__()
        self.setMouseTracking(True)
        self.grid = grid
        self.pos = None
        self.zoom = 1
        #
        self.initUI()

    def initUI(self):
        self.show()

    def paintEvent(self, paint_event):
        painter = QPainter(self)
        super().paintImage(painter)
        painter.end()

    def switch_imgVis(self):
        super().make_rgb_img(self.grid.imgs.get("crop"))
        self.repaint()
        self.updateMag()

    def switch_imgK(self):
        super().make_idx8_img(self.grid.imgs.get("kmean"), self.grid.imgs.getParam("k"))
        self.repaint()
        self.updateMag()

    def switch_imgB(self):
        super().make_bin_img(self.grid.imgs.get("bin"))
        self.repaint()
        self.updateMag()

    def mouseMoveEvent(self, event):
        self.updateMag()
        self.repaint()

    def mousePressEvent(self, event):
        if event.button() == Qt.RightButton:
            self.zoom = (self.zoom + 1) % 3
            self.mouseMoveEvent(event)

    def updateMag(self):
        pos = self.mapFromGlobal(QCursor().pos())
        if self.zoom != 0:
            magnifying_glass(
                self, pos, area=int(self.width() / 7), zoom=self.zoom * 1.5
            )
        else:
            self.setCursor(QCursor(Qt.ArrowCursor))



# basic imports
import numpy as np
import pandas as pd
import sys

# self imports
from grid.io import *
from grid.lib import *


class GMap:
    """ """

    def __init__(self):
        """
        ----------
        Parameters
        ----------
        """

        # constant var
        self._degRot = range(-75, 90 + 1, 15)

        # map file
        self.pdMap = None
        self.isMap = False
        self.countNames = 0

        # dimension
        self.nAxs = [0, 0]
        self.nAxsCur = [0, 0]
        self.nRow = 0
        self.nCol = 0
        self.img = None
        self.imgBin = None
        self.imgRGB = None
        self.imgH = 0
        self.imgW = 0

        # axes
        self.angles = [0, 0]
        self.anglesCur = [0, 0]
        self.slps = [0, 0]  # 0 and 90 degrees for common images
        self.sigs = [0, 0]
        self.itcs = [0, 0]
        self.slpsReset = [0, 0]
        self.itcsReset = [0, 0]

        # output
        self.dt = None

        # progress bar
        self.flag = True
        self.subflag = True
        self.window = 1000

    def load(self, pathMap):
        """
        ----------
        Parameters
        ----------
        """

        self.pdMap = loadMap(pathMap)

        # if self.pdMap is not None:
        #     self.nRow, self.nCol = self.pdMap.shape
        #     self.isMap = True

    def findPlots(self, imgRGB, imgBin, nRow=0, nCol=0, nSmooth=100):
        """
        ----------
        Parameters
        ----------
        """

        # get image info
        self.imgBin = imgBin
        self.imgRGB = imgRGB
        self.imgH, self.imgW = imgBin.shape[:2]
        isDimAssigned = nRow != 0 and nCol != 0

        # if dim is assigned regardless have map or not, force changning nR and nC
        if isDimAssigned:
            self.nAxs = [nCol, nRow]
        elif self.pdMap is not None:
            self.nAxs = [self.pdMap.shape[1], self.pdMap.shape[0]]

        # find angles and slopes
        self.angles = self.detectAngles(img=imgBin, rangeAngle=self._degRot)

        # find intercepts and make centers as pandas DT
        self.locateCenters(nSmooth)

    def detectAngles(self, img, rangeAngle):
        # evaluate each angle
        sc = []
        for angle in rangeAngle:
            imgR = rotateBinNdArray(img, angle)
            sig = imgR.mean(axis=0)
            sigFour = getFourierTransform(sig)
            sc.append(max(sigFour))

        # angle with maximum score win
        scSort = sc.copy()
        scSort.sort()
        idxMax = [i for i in range(len(sc)) if (sc[i] in scSort[-2:])]
        angles = np.array([rangeAngle[idx] for idx in idxMax])

        # sort angles (one closer to 0/90 is major angle)
        idx_major = np.argmin(angles % 90)
        angles = angles[[idx_major, abs(idx_major - 1)]]

        # return
        return angles

    def locateCenters(self, nSmooth=100):
        self.slps = [1 / np.tan(np.pi / 180 * angle) for angle in self.angles]

        # progress bar
        prog = None
        if self.flag:
            self.flag = False
            prog = initProgress(2, "Calculate slopes and intercepts")

        # find intercepts given 2 angles
        self.updateIntercepts(self.angles, self.nAxs, nSmooth)

        # get pandas data table
        updateProgress(prog, flag=self.subflag)
        self.dt = self.getDfCoordinate(self.angles, self.slps, self.itcs)

        # end progress bar
        if self.subflag:
            self.subflag = False
            QTimer.singleShot(self.window, lambda: setattr(self, "flag", True))
            QTimer.singleShot(self.window, lambda: setattr(self, "subflag", True))

    def updateIntercepts(self, angles, nSigs, nSmooth):
        for i in range(2):
            if (
                nSigs[i] == 0
                or self.nAxsCur[i] != nSigs[i]
                or self.angles[i] != self.anglesCur[i]
            ):
                sig, intercept = self.cpuIntercept(angles[i], nSigs[i], nSmooth)

                self.sigs[i] = sig
                self.itcs[i] = intercept

                # update number of peaks
                self.nAxs[i] = len(intercept)
                self.nAxsCur[i] = len(intercept)

                # update angles
                self.anglesCur[i] = self.angles[i]

        # self.imgsR_Bin[i] = imgR_bin
        # self.imgsR_RGB[i] = imgR_rgb
        # self.imgHr[i] = imgR_bin.shape[0]
        # self.imgWr[i] = imgR_bin.shape[1]

    def cpuIntercept(self, angle, nSig, nSmooth):
        imgR_bin = rotateBinNdArray(self.imgBin, angle)
        sig = findPeaks(img=imgR_bin, nPeaks=nSig, nSmooth=nSmooth)[0]
        intercept = getCardIntercept(sig, angle, self.imgH)
        if isinstance(intercept, int):
            intercept = list([intercept])
        return sig, intercept

    def getDfCoordinate(self, angles, slopes, intercepts):
        """
        ----------
        Parameters
        ----------
        """
        imgH, imgW = self.imgBin.shape
        tol = 0.025
        bdN, bdS = -imgH * tol, imgH * (1 + tol)
        bdW, bdE = -imgW * tol, imgW * (1 + tol)

        idxCol = 0 if abs(slopes[0]) > abs(slopes[1]) else 1
        idxRow = 1 - idxCol
        idxMaj = 0 if getClosedTo0or90(angles[0]) <= getClosedTo0or90(angles[1]) else 1
        idxMin = 1 - idxMaj

        itc_maj = np.sort(intercepts[0])
        itc_min = np.sort(intercepts[1])

        plotsMaj = []
        plotsMin = []
        pts = []

        pMaj = 0
        for itcMaj in itc_maj:  # major
            pMin = 0
            for itcMin in itc_min:  # minor
                ptX, ptY = solveLines(slopes[idxMaj], itcMaj, slopes[idxMin], itcMin)
                if ptX >= bdW and ptX <= bdE and ptY >= bdN and ptY <= bdS:
                    # outside the frame but within tolerant range
                    if ptX < 0:
                        ptX = 0
                    elif ptX >= imgW:
                        ptX = imgW - 1
                    if ptY < 0:
                        ptY = 0
                    elif ptY >= imgH:
                        ptY = imgH - 1
                    plotsMaj.append(pMaj)
                    plotsMin.append(pMin)
                    pts.append((ptX, ptY))
                    pMin += 1
            pMaj += 1

        # col order would be flipped if the angle is greater than zero
        if angles[1] > 0 and angles[1] < 90 and angles[0] != 0:
            plotsMin = plotsMin[::-1]

        if idxCol == idxMaj:
            dataframe = pd.DataFrame({"row": plotsMin, "col": plotsMaj, "pt": pts})
        elif idxCol == idxMin:
            dataframe = pd.DataFrame({"row": plotsMaj, "col": plotsMin, "pt": pts})

        self.nRow, self.nCol = dataframe["row"].max() + 1, dataframe["col"].max() + 1

        # update map names
        ctNA = 0
        names = []
        for _, entry in dataframe.iterrows():
            row = entry.row
            col = entry.col
            try:
                names.append(self.pdMap.iloc[row, col])
            except Exception:
                names.append("unnamed_%d" % (ctNA))
                ctNA += 1
        dataframe["var"] = names

        # return
        return dataframe

    def getCoordinate(self, row, col):
        """
        ----------
        Parameters
        ----------
        """
        dt = self.dt
        return dt[(dt.row == row) & (dt.col == col)]["pt"].values[0]

    def getName(self, row, col):
        """
        ----------
        Parameters
        ----------
        """
        dt = self.dt
        return dt[(dt.row == row) & (dt.col == col)]["var"].values[0]

    def delAnchor(self, axis, index):
        # signals and interecepts
        sigMaj = np.array(self.sigs[0])
        sigMin = np.array(self.sigs[1])

        # handle error (and I don't know why)
        if index >= len(self.sigs[axis]):
            index = len(self.sigs[axis]) - 1

        if axis == 0:
            sigMaj = np.delete(sigMaj, index)
            itcMaj = getCardIntercept(self.sigs[axis], self.angles[axis], self.imgH)
            itcMin = self.itcs[1]
        else:
            sigMin = np.delete(sigMin, index)
            itcMaj = self.itcs[0]
            itcMin = getCardIntercept(self.sigs[axis], self.angles[axis], self.imgH)

        self.sigs = [sigMaj, sigMin]
        self.itcs = [itcMaj, itcMin]

        # update dataframe
        self.dt = self.getDfCoordinate(self.angles, self.slps, self.itcs)

    def addMajAnchor(self, value):
        # update signals on major axis
        sigMaj = np.array(self.sigs[0])
        sigMaj = np.append(sigMaj, value)
        sigMin = np.array(self.sigs[1])
        self.sigs = [sigMaj, sigMin]

        # update intercept on major axis
        itcMaj = getCardIntercept(self.sigs[0], self.angles[0], self.imgH)
        itcMin = np.array(self.itcs[1])
        self.itcs = [itcMaj, itcMin]

        # update dataframe
        self.dt = self.getDfCoordinate(self.angles, self.slps, self.itcs)

    def addMinAnchor(self, itc):
        # signal
        new_signal = getSigFromItc(itc, self.angles[1], self.imgH)
        sigMaj = np.array(self.sigs[0])
        sigMin = np.array(self.sigs[1])
        sigMin = np.append(sigMin, new_signal)
        self.sigs = [sigMaj, sigMin]

        # intercepts
        itcMaj = np.array(self.itcs[0])
        itcMin = np.array(self.itcs[1])
        itcMin = np.append(itcMin, itc)
        self.itcs = [itcMaj, itcMin]

        # update dataframe
        self.dt = self.getDfCoordinate(self.angles, self.slps, self.itcs)

    def modMajAnchor(self, index, value):
        # signals
        self.sigs[0][index] = value

        # intercepts
        itcMaj = getCardIntercept(self.sigs[0], self.angles[0], self.imgH)
        itcMin = np.array(self.itcs[1])
        self.itcs = [itcMaj, itcMin]

        # dataframe
        self.dt = self.getDfCoordinate(self.angles, self.slps, self.itcs)

    def modMinAnchor(self, index, itc):
        # angles
        angle = self.angles[1]

        # signals
        self.sigs[1][index] = getSigFromItc(itc, angle, self.imgH)

        # intercept
        itcMaj = np.array(self.itcs[0])
        itcMin = getCardIntercept(self.sigs[1], self.angles[1], self.imgH)
        self.itcs = [itcMaj, itcMin]

        # dataframe
        self.dt = self.getDfCoordinate(self.angles, self.slps, self.itcs)


def getClosedTo0or90(x):
    s1 = abs(abs(x) - 90)
    s2 = abs(abs(x))
    return min(s1, s2)


def scaleTo0and1(array, length):
    array = np.array(array)
    return array / length


def scaleToOrg(array, length):
    array = np.array(array)
    return array * length



# basic imports
import numpy as np
import shapefile

# self imports
from grid.io import *
from grid.lib import *


class GImage:
    """ """

    def __init__(self):
        """
        ----------
        Parameters
        ----------
        """

        # images
        self.imgs = {
            "raw": None,
            "rawRs": None,
            "crop": None,
            "mean": None,
            "kmean": None,
            "binOrg": None,
            "binTemp": None,
            "binSd": None,
            "binSm": None,
            "bin": None,
            "binSeg": None,
            "visSeg": None,
        }

        # dimension
        self.width, self.height, self.depth = 0, 0, 0
        self.widthRs, self.heightRs = 0, 0
        self.shape = (self.height, self.width, self.depth)
        self.shapeRs = (self.heightRs, self.widthRs, self.depth)
        self.n_rot = 0

        # crop/ geo-reference
        self.f_shp = None
        self.hasShp = False
        self.tiff_transform = None
        self.crs = False
        self.pts_crop = []
        self.mat_H = None

        # kmean param.
        self.paramKMs = {
            "k": -1,
            "center": None,
            "rank": None,
            "features": [],
            "lsSelect": [],
            "lsKTobin": [],
            "valShad": -1,
            "valSmth": -1,
        }

    def get(self, key):
        """
        ----------
        Parameters
        ----------
        """

        return self.imgs[key]

    def set(self, key, value):
        """
        ----------
        Parameters
        ----------
        """

        self.imgs[key] = value

    def load(self, pathImg, pathShp=None):
        """
        ----------
        Parameters
        ----------
        """
        if isinstance(pathImg, str):
            # a path to a file
            isLocalImg = pathImg.find("http://") + pathImg.find("https://") == -2

            # image
            if isLocalImg:
                imgInput, self.tiff_transform, self.crs = loadImg(pathImg)
                try:
                    self.f_shp = shapefile.Reader(pathShp)
                    tmp = self.f_shp.shapeRecords()  # test if it can be read
                    self.hasShp = True
                except Exception:
                    self.f_shp = None
                    self.hasShp = False
            else:
                imgInput = loadImgWeb(pathImg)
        elif isinstance(pathImg, np.ndarray):
            # is a numpy
            imgInput = pathImg

        # assign
        self.set(key="raw", value=imgInput)
        self.setShape(shape=imgInput.shape, isRaw=True)

    def getParam(self, key):
        """
        ----------
        Parameters
        ----------
        """

        return self.paramKMs[key]

    def setParam(self, key, value):
        """
        ----------
        Parameters
        ----------
        """

        self.paramKMs[key] = value

    def resetParam(self):
        """
        ----------
        Parameters
        ----------
        """

        self.paramKMs = {
            "k": -1,
            "center": None,
            "features": [],
            "lsSelect": [],
            "lsKTobin": [],
            "valShad": -1,
            "valSmth": -1,
        }

    def crop(self, pts=None):
        """
        ----------
        Parameters
        ----------
        """

        if pts is None:
            # (x, y) = (W, H)
            pts = np.float32(
                [
                    [0, 0],
                    [self.shape[1], 0],
                    [self.shape[1], self.shape[0]],
                    [0, self.shape[0]],
                ]
            )

        imgCrop, M = cropImg(self.imgs["raw"], pts)
        self.set(key="crop", value=imgCrop)
        self.set(key="mean", value=self.get("crop")[:, :, :3].mean(axis=2))
        self.setShape(shape=self.get(key="crop").shape)
        self.mat_H = M
        self.pts_crop = pts.copy()

        self.resetParam()

    def doKMeans(self, k=3, features=[0, 1, 2], colorOnly=False):
        """
        ----------
        Parameters
        ----------
        """

        if len(features) == 0:
            print("no feature is selected")
        elif (k != self.paramKMs["k"]) or (features != self.paramKMs["features"]):
            # Will skip if no updates on the params
            imgK, center = doKMeans(img=self.get("crop"), k=k, features=features)
            self.set(key="kmean", value=imgK)
            # update parameters
            self.paramKMs["center"] = center
            self.paramKMs["rank"] = self.rankCenters(
                k=k, center=center, colorOnly=colorOnly
            )
            # can't update yet as binarize needs it
            # self.paramKMs['k'] = k
            # self.paramKMs['features'] = features
        else:
            # skip
            bugmsg("skip kmean")

    def binarize(self, k=3, features=[0, 1, 2], lsSelect=[0]):
        """
        ----------
        Parameters
        ----------
        """

        if len(features) == 0:
            print("no feature is selected")
        elif (
            (k != self.paramKMs["k"])
            or (features != self.paramKMs["features"])
            or (lsSelect != self.paramKMs["lsSelect"])
        ):
            # ratioK = [(self.paramKMs['center'][i, 0]-self.paramKMs['center'][i, 1])/self.paramKMs['center'][i, :].sum()
            #           for i in range(self.paramKMs['center'].shape[0])]
            # # rankK = np.flip(np.argsort(ratioK), 0)
            # rankK = np.argsort(ratioK)
            try:
                clusterSelected = self.paramKMs["rank"][lsSelect]
            except Exception:
                clusterSelected = []
            self.set(
                key="binOrg",
                value=((np.isin(self.get("kmean"), clusterSelected)) * 1).astype(int),
            )
            self.set(key="binTemp", value=self.get("binOrg").copy())
            self.set(key="binSm", value=self.get("binOrg").copy())
            # udpate parameters
            self.paramKMs["k"] = k
            self.paramKMs["features"] = features
            self.paramKMs["lsSelect"] = lsSelect
            self.paramKMs["lsKToBin"] = clusterSelected
        else:
            # skip
            bugmsg("skip binarize")

    def smooth(self, value):
        """
        ----------
        Parameters
        ----------
        """

        if value != self.paramKMs["valSmth"]:
            valSmthDiff = value - self.paramKMs["valSmth"]
            if valSmthDiff > 0:
                valSmthReal = valSmthDiff
            else:
                valSmthReal = value
                self.set(key="binTemp", value=self.get(key="binOrg").copy())
            self.set(
                key="binTemp",
                value=smoothImg(image=self.get(key="binTemp"), n=valSmthReal),
            )
            self.set(key="binSm", value=binarizeSmImg(self.get(key="binTemp")))
            # update parameters
            self.paramKMs["valSmth"] = value
        else:
            # skip
            bugmsg("skip smoothing")

    def deShadow(self, value):
        """
        ----------
        Parameters
        ----------
        """

        if value != self.paramKMs["valShad"]:
            self.set(key="binSd", value=(self.get(key="mean") >= value) * 1)
            # update parameter
            self.paramKMs["valShad"] = value
        else:
            # skip
            bugmsg("skip shadow")

    def finalized(self):
        """
        ----------
        Parameters
        ----------
        """

        self.set(key="bin", value=np.multiply(self.get("binSm"), self.get("binSd")))

    def readyForSeg(self):
        """
        ----------
        Parameters
        ----------
        """
        nSmt = int(min(self.get("bin").shape[0], self.get("bin").shape[1]) / 300)
        self.set(key="binSeg", value=blurImg(self.get("bin"), n=nSmt))
        # compute the vis/seg image
        imgTemp = np.expand_dims(self.get("bin"), axis=2)
        imgSeg = np.multiply(self.get("crop")[:, :, :3], imgTemp).copy()
        imgSeg[(imgSeg.mean(axis=2) == 0), :] = 255
        self.set(key="visSeg", value=imgSeg)

    def rankCenters(self, k, center, colorOnly=False):
        scores = []

        if colorOnly:
            ratioK = [
                (center[i, 0] - center[i, 1]) / center[i, :].sum()
                for i in range(center.shape[0])
            ]
            rank = np.flip(np.argsort(ratioK), axis=0)
        else:
            for i in range(k):
                imgB = (np.isin(self.get("kmean"), i) * 1).astype(int)
                sigs = imgB.mean(axis=0)
                sigsF = getFourierTransform(sigs)
                scMaxF = round((max(sigsF) / sigsF.mean()) / 100, 4)  # [0, 1]
                scMean = round(1 - (sigs.mean()), 4)  # [0, 1]
                try:
                    scPeaks = round(
                        len(find_peaks(sigs, height=(sigs.mean()))[0])
                        / len(find_peaks(sigs)[0]),
                        4,
                    )
                except Exception:
                    scPeaks = 0

                score = scMaxF * 0.25 + scMean * 0.25 + scPeaks * 0.5
                scores.append(score)
            rank = np.flip(np.array(scores).argsort(), axis=0)

        return rank

    def rotate(self, nRot):
        """
        ----------
        Parameters
        ----------
        """

        self.n_rot += nRot

        for key in self.imgs.keys():
            if key == "raw" or key == "rawRs":
                # only rotate cropped images
                continue
            try:
                self.set(key=key, value=np.rot90(self.get(key=key), nRot))
            except Exception:
                None
        self.setShape(self.get("crop").shape)

    def setShape(self, shape, isRaw=False):
        """
        ----------
        Parameters
        ----------
        """
        if isRaw:
            try:
                self.height, self.width, self.depth = shape
            except Exception:
                self.height, self.width = shape
                self.depth = 1
            self.shape = (self.height, self.width, self.depth)
        else:
            try:
                self.heightRs, self.widthRs, self.depth = shape
            except Exception:
                self.heightRs, self.widthRs = shape
                self.depth = 1
            self.shapeRs = (self.heightRs, self.widthRs, self.depth)

    def getShape(self, is3D=False, isRaw=False):
        """
        ----------
        Parameters
        ----------
        """
        if is3D:
            return self.shape if isRaw else self.shapeRs
        else:
            return self.shape[:2] if isRaw else self.shapeRs[:2]



# basic imports
import numpy as np
import pandas as pd
import os

# 3rd party imports
from PyQt6.QtWidgets import *
from PyQt6.QtCore import *
from PyQt6.QtGui import *

# self imports
from grid.grid import *
from grid.io import *
from grid.gui.customQt import *


class PnOutputer(QWidget):
    def __init__(self, grid):
        """ """
        super().__init__()
        self.setFocusPolicy(Qt.StrongFocus)
        self.setFocus()
        self.update()
        self.grid = grid
        self.runDefaultSeg()

        self.layout = QHBoxLayout()
        """left side"""
        self.wg_img = Widget_Seg(grid)
        """right side"""
        self.pn_right = QWidget()
        self.lo_right = QVBoxLayout()
        self.sc_right = QScrollArea()
        self.sc_right.setWidgetResizable(True)
        # Boundary
        self.gr_seg = QGroupBox("Segmentation")
        self.lo_seg = QGridLayout()
        # Boundary (auto)
        self.gr_auto = QGroupBox("Dynamic")
        self.lo_auto = QVBoxLayout()
        self.gr_grid = QGroupBox("Grid Coef. = 0.0")
        self.lo_grid = QVBoxLayout()
        self.sl_grid = QSlider(Qt.Horizontal)
        # Boundary (fix)
        self.gr_fix = QGroupBox("Fixed")
        self.lo_fix = QVBoxLayout()
        self.gr_width = QGroupBox("Width = 50 units")
        self.lo_width = QVBoxLayout()
        self.sl_width = QSlider(Qt.Horizontal)
        self.gr_length = QGroupBox("Length = 50 units")
        self.lo_length = QVBoxLayout()
        self.sl_length = QSlider(Qt.Horizontal)
        self.lb_alignX = QLabel("Align Columns (beta)")
        self.cb_alignX = QComboBox()
        self.lb_alignY = QLabel("Align Rows (beta)")
        self.cb_alignY = QComboBox()
        self.ck_evenH = QCheckBox("Evenly Distribute Columns (beta)")
        self.ck_evenV = QCheckBox("Evenly Distribute Rows (beta)")
        self.bt_reset = QPushButton("Reset")
        # Tool
        self.gr_tol = QGroupBox("Tools (Right-click to switch)")
        self.lo_tol = QGridLayout()
        self.rb_ct = QRadioButton("Adjust Centroid")
        self.rb_adj = QRadioButton("Adjust Border")
        self.rb_vp = QRadioButton("Pan (Vertical)")
        self.rb_hp = QRadioButton("Pan (Horizontal)")
        # Display
        self.gr_dis = QGroupBox("Display")
        self.lo_dis = QHBoxLayout()
        self.rb_srgb = QRadioButton("Selected RGB (A)")
        self.rb_rgb = QRadioButton("RGB (S)")
        # Output
        self.gr_out = QGroupBox("Output")
        self.lo_out = QGridLayout()
        self.lb_project = QLabel("Prefix")
        self.fd_project = QLineEdit("GRID")
        self.lb_output = QLabel("Output Path")
        self.fd_output = QLineEdit(self.grid.path_out)
        self.bt_output = QPushButton("Browse")
        self.ck_simple = QCheckBox("Simple output")
        """ui"""
        self.initUI()

    def initUI(self):
        """seg-auto (right)"""
        # components
        self.sl_grid.setMinimum(0)
        self.sl_grid.setMaximum(10)
        self.sl_grid.setValue(0)
        self.sl_grid.setTickInterval(2)
        self.sl_grid.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.sl_grid.valueChanged.connect(self.change_grid)
        self.gr_auto.setCheckable(True)
        self.gr_auto.setChecked(True)
        self.gr_auto.clicked.connect(self.auto_seg)
        # layout
        self.lo_grid.addWidget(self.sl_grid)
        self.gr_grid.setLayout(self.lo_grid)
        self.lo_auto.addWidget(self.gr_grid)
        self.gr_auto.setLayout(self.lo_auto)
        """seg-fix (right)"""
        # components
        self.sl_width.setMinimum(0)
        self.sl_width.setMaximum(100)
        self.sl_width.setValue(50)
        self.sl_width.setTickInterval(2)
        self.sl_width.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.sl_width.valueChanged.connect(self.change_width)
        self.sl_length.setMinimum(0)
        self.sl_length.setMaximum(100)
        self.sl_length.setValue(50)
        self.sl_length.setTickInterval(2)
        self.sl_length.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.sl_length.valueChanged.connect(self.change_length)
        self.cb_alignX.addItem("None")
        self.cb_alignX.addItem("Left")
        self.cb_alignX.addItem("Center")
        self.cb_alignX.addItem("Right")
        self.cb_alignY.addItem("None")
        self.cb_alignY.addItem("Top")
        self.cb_alignY.addItem("Middle")
        self.cb_alignY.addItem("Bottom")
        self.cb_alignX.currentIndexChanged.connect(self.alignX)
        self.cb_alignY.currentIndexChanged.connect(self.alignY)
        self.ck_evenH.clicked.connect(self.evenH)
        self.ck_evenV.clicked.connect(self.evenV)
        self.bt_reset.clicked.connect(self.reset)
        self.gr_fix.setCheckable(True)
        self.gr_fix.setChecked(False)
        self.gr_fix.clicked.connect(self.fix_seg)
        # layout
        self.lo_width.addWidget(self.sl_width)
        self.gr_width.setLayout(self.lo_width)
        self.lo_length.addWidget(self.sl_length)
        self.gr_length.setLayout(self.lo_length)
        # self.gr_aln.setLayout(self.lo_aln)
        self.lo_fix.addWidget(self.gr_width)
        self.lo_fix.addWidget(self.gr_length)
        self.lo_fix.addWidget(self.lb_alignX)
        self.lo_fix.addWidget(self.cb_alignX)
        self.lo_fix.addWidget(self.lb_alignY)
        self.lo_fix.addWidget(self.cb_alignY)
        self.lo_fix.addWidget(self.ck_evenH)
        self.lo_fix.addWidget(self.ck_evenV)
        self.gr_fix.setLayout(self.lo_fix)
        """seg (right)"""
        # layout
        self.lo_seg.addWidget(self.gr_auto)
        self.lo_seg.addWidget(self.gr_fix)
        self.gr_seg.setLayout(self.lo_seg)
        """tool"""
        self.rb_adj.setChecked(True)
        self.rb_ct.toggled.connect(lambda: self.changeTool(index=0))
        self.rb_adj.toggled.connect(lambda: self.changeTool(index=1))
        self.rb_vp.toggled.connect(lambda: self.changeTool(index=2))
        self.rb_hp.toggled.connect(lambda: self.changeTool(index=3))
        self.lo_tol.addWidget(self.rb_ct, 0, 0)
        self.lo_tol.addWidget(self.rb_adj, 0, 1)
        self.lo_tol.addWidget(self.rb_vp, 1, 0)
        self.lo_tol.addWidget(self.rb_hp, 1, 1)
        self.gr_tol.setLayout(self.lo_tol)
        """display (right)"""
        # components
        self.rb_srgb.setChecked(True)
        self.rb_srgb.toggled.connect(self.wg_img.switch_imgSVis)
        self.rb_rgb.toggled.connect(self.wg_img.switch_imgVis)
        # layout
        self.lo_dis.addWidget(self.rb_srgb)
        self.lo_dis.addWidget(self.rb_rgb)
        self.gr_dis.setLayout(self.lo_dis)
        """output (right)"""
        # components
        font = self.fd_project.font()
        font.setPointSize(25)
        fm = QFontMetrics(font)
        self.fd_project.setFixedHeight(fm.height())
        self.fd_output.setFixedHeight(fm.height())
        self.bt_output.clicked.connect(self.assign_PathOut)
        self.ck_simple.setChecked(True)
        # layout
        self.lo_out.addWidget(self.lb_project, 0, 0)
        self.lo_out.addWidget(self.fd_project, 0, 1)
        self.lo_out.addWidget(self.lb_output, 1, 0)
        self.lo_out.addWidget(self.fd_output, 1, 1)
        self.lo_out.addWidget(self.bt_output, 1, 2)
        self.lo_out.addWidget(self.ck_simple, 2, 0, 1, 3)
        self.gr_out.setLayout(self.lo_out)
        """layout"""
        # left
        # NONE
        # right
        self.lo_right.addWidget(self.gr_seg)
        self.lo_right.addWidget(self.gr_tol)
        self.lo_right.addWidget(self.bt_reset)
        self.lo_right.addWidget(self.gr_dis)
        self.lo_right.addWidget(self.gr_out)
        self.pn_right.setLayout(self.lo_right)
        self.sc_right.setWidget(self.pn_right)
        # policy
        policy_right = QSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )
        policy_right.setHorizontalStretch(2)
        self.sc_right.setSizePolicy(policy_right)
        policy_left = QSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )
        policy_left.setHorizontalStretch(3)
        self.wg_img.setSizePolicy(policy_left)
        # assemble
        self.layout.addWidget(self.wg_img)
        self.layout.addWidget(self.sc_right)
        self.setLayout(self.layout)
        # collapsable
        self.collapse(isAuto=True)
        # show
        self.show()

    def collapse(self, isAuto=True):
        self.gr_grid.setVisible(isAuto)
        self.gr_width.setVisible(not isAuto)
        self.gr_length.setVisible(not isAuto)
        self.lb_alignX.setVisible(not isAuto)
        self.cb_alignX.setVisible(not isAuto)
        self.lb_alignY.setVisible(not isAuto)
        self.cb_alignY.setVisible(not isAuto)
        self.ck_evenH.setVisible(not isAuto)
        self.ck_evenV.setVisible(not isAuto)

    def auto_seg(self):
        """ """
        self.gr_auto.setChecked(True)
        self.gr_fix.setChecked(False)
        self.collapse(isAuto=True)
        val_grid = 1 - (self.sl_grid.value() / 10)
        self.wg_img.auto_seg(coefGrid=val_grid)

    def fix_seg(self):
        """ """
        self.gr_auto.setChecked(False)
        self.gr_fix.setChecked(True)
        self.collapse(isAuto=False)
        value_width = self.sl_width.value()
        value_length = self.sl_length.value()
        self.wg_img.fix_seg(value_width, value_length)

    def changeTool(self, index):
        self.wg_img.task = index

    def change_grid(self):
        """ """
        value = self.sl_grid.value()
        self.gr_grid.setTitle("Grid Coef. = %.2f" % (value / 10))
        self.auto_seg()

    def change_width(self):
        """ """
        value = self.sl_width.value()
        self.gr_width.setTitle("Width = %d units" % (value))
        self.fix_seg()

    def change_length(self):
        """ """
        value = self.sl_length.value()
        self.gr_length.setTitle("Length = %d units" % (value))
        self.fix_seg()

    def alignX(self):
        """ """
        index = self.cb_alignX.currentIndex()
        self.wg_img.align(method=index, axis=1)

    def alignY(self):
        """ """
        index = self.cb_alignY.currentIndex()
        self.wg_img.align(method=index, axis=0)

    def evenH(self):
        """ """
        self.wg_img.distributed(axis=1, isEven=self.ck_evenH.isChecked())

    def evenV(self):
        """ """
        self.wg_img.distributed(axis=0, isEven=self.ck_evenV.isChecked())

    def reset(self):
        """ """
        self.sl_width.setValue(50)
        self.sl_length.setValue(50)
        self.cb_alignX.setCurrentIndex(0)
        self.cb_alignY.setCurrentIndex(0)
        self.grid.agents.resetCoordinate()
        self.runDefaultSeg()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_A:
            self.rb_srgb.setChecked(True)
        elif event.key() == Qt.Key_S:
            self.rb_rgb.setChecked(True)

    def assign_PathOut(self):
        path = QFileDialog().getExistingDirectory(
            self,
            "",
            "",
            QFileDialog.Option.ShowDirsOnly,
        )
        self.fd_output.setText(path)

    def paintEvent(self, paint_event):
        if self.wg_img.task == 0:
            self.rb_ct.setChecked(True)
        elif self.wg_img.task == 1:
            self.rb_adj.setChecked(True)
        elif self.wg_img.task == 2:
            self.rb_vp.setChecked(True)
        else:
            self.rb_hp.setChecked(True)

    def runDefaultSeg(self):
        if self.grid.imgs.hasShp:
            self.grid.imgs.readyForSeg()
            self.grid.agents.setup(gmap=self.grid.map, gimg=self.grid.imgs)
        else:
            self.grid.cpuSeg()


class Widget_Seg(Widget_Img):
    def __init__(self, grid):
        """ """
        super().__init__()
        self.setMouseTracking(True)
        self.grid = grid
        self.img_raw = grid.imgs.get("visSeg")
        self.task = 0  # 0 centroid, 1 zoom, 2 panV, 3 panH
        """attr"""
        # painter
        self.ratio = 0
        # mouse
        self.agent_click = False
        self.dir = None
        # seg
        imgBin = self.grid.imgs.get("bin")
        imgBinTemp = imgBin.reshape(imgBin.shape[0], imgBin.shape[1], 1)
        self.img_seg = np.multiply(
            self.grid.imgs.get("crop")[:, :, :3], imgBinTemp
        ).copy()

        # ui
        self.initUI()

    def initUI(self):
        self.make_rgb_img(self.img_raw)
        self.show()

    def mouseReleaseEvent(self, event):
        self.agent_click = False

    def mousePressEvent(self, event):
        self.pos_press = (event.pos().x(), event.pos().y())
        self.pos_move_prev = self.pos_press

        # to figure out what border has benn selected
        for row in range(self.grid.map.nRow):
            for col in range(self.grid.map.nCol):
                # compute rect of agent
                agent = self.grid.agents.get(row, col)
                if not agent or agent.isFake():
                    continue
                rect = agent.getQRect()
                if self.isFitWidth:
                    self.ratio = self.width() / self.qimg.width()
                else:
                    self.ratio = self.height() / self.qimg.height()

                rec_agent = QRect(
                    int(rect.x() * self.ratio + self.rgX[0]),
                    int(rect.y() * self.ratio + self.rgY[0]),
                    int(rect.width() * self.ratio),
                    int(rect.height() * self.ratio),
                )
                # if contain cursor
                if rec_agent.contains(event.pos()):
                    self.agent_click = agent
                    if self.task == 1:
                        # zoom and mod border
                        bd_W = rec_agent.x()
                        bd_N = rec_agent.y()
                        bd_E = bd_W + rec_agent.width()
                        bd_S = bd_N + rec_agent.height()
                        dis_W = abs(self.pos_press[0] - bd_W)
                        dis_N = abs(self.pos_press[1] - bd_N)
                        dis_E = abs(self.pos_press[0] - bd_E)
                        dis_S = abs(self.pos_press[1] - bd_S)
                        dir_idx = np.argmin(np.array([dis_N, dis_W, dis_S, dis_E]))
                        if dir_idx == 0:
                            self.dir = Dir.NORTH
                        elif dir_idx == 1:
                            self.dir = Dir.WEST
                        elif dir_idx == 2:
                            self.dir = Dir.SOUTH
                        elif dir_idx == 3:
                            self.dir = Dir.EAST
                    break

        # mag module
        if event.button() == Qt.RightButton:
            self.task = (self.task + 1) % 4
            self.mouseMoveEvent(event)

    def mouseMoveEvent(self, event):
        self.pos_move = (event.pos().x(), event.pos().y())

        # change cursor
        if self.task == 0:
            # moving centroid
            self.setCursor(QCursor(Qt.SizeAllCursor))
        elif self.task == 1:
            # adjust borders
            magnifying_glass(self, event.pos(), area=int(self.width() / 7), zoom=1.5)
        elif self.task == 2:
            self.setCursor(QCursor(Qt.SizeVerCursor))
        elif self.task == 3:
            self.setCursor(QCursor(Qt.SizeHorCursor))

        # pan
        if (event.buttons() == Qt.LeftButton) & (self.agent_click != False):
            # convert GUI to image coordinate
            x_move, y_move = self.convertGUI2XY(self.pos_move)
            x_move_prev, y_move_prev = self.convertGUI2XY(self.pos_move_prev)
            dx = x_move - x_move_prev
            dy = y_move - y_move_prev

            if self.task == 0:
                # adjust centroid
                self.agent_click.updateCoordinate(value=dy, axis=0)
                self.agent_click.updateCoordinate(value=dx, axis=1)
            elif self.task == 1:
                # adjust border
                if self.dir == Dir.NORTH or self.dir == Dir.SOUTH:
                    value = y_move
                elif self.dir == Dir.WEST or self.dir == Dir.EAST:
                    value = x_move
                self.grid.agents.setBorder(self.agent_click, self.dir, value)
            elif self.task == 2:
                # V pan
                row = self.agent_click.row
                self.grid.agents.pan(axis=0, target=row, value=dy)
            elif self.task == 3:
                # H pan
                col = self.agent_click.col
                self.grid.agents.pan(axis=1, target=col, value=dx)

        self.pos_move_prev = self.pos_move
        self.repaint()

    def convertGUI2XY(self, pt):
        posX = (pt[0] - self.rgX[0]) / self.ratio
        posY = (pt[1] - self.rgY[0]) / self.ratio
        return (posX, posY)

    def paintEvent(self, paint_event):
        painter = QPainter(self)
        super().paintImage(painter)
        pen = QPen()
        pen.setWidth(3)
        pen.setColor(Qt.red)
        painter.setPen(pen)
        painter.setBrush(Qt.transparent)
        for row in range(self.grid.agents.nRow):
            for col in range(self.grid.agents.nCol):
                agent = self.grid.agents.get(row, col)
                if not agent or agent.isFake():
                    continue
                rect = agent.getQRect()
                pt_x, pt_y = agent.getCoordinate()
                self.ratio = (
                    self.width() / self.qimg.width()
                    if self.isFitWidth
                    else self.height() / self.qimg.height()
                )
                rec_agent = QRect(
                    int(rect.x() * self.ratio + self.rgX[0]),
                    int(rect.y() * self.ratio + self.rgY[0]),
                    int(rect.width() * self.ratio),
                    int(rect.height() * self.ratio),
                )
                drawCross(
                    pt_x * self.ratio + self.rgX[0],
                    pt_y * self.ratio + self.rgY[0],
                    painter,
                )
                painter.drawRect(rec_agent)
        painter.end()

    def switch_imgVis(self):
        super().make_rgb_img(self.grid.imgs.get("crop"))
        self.repaint()

    def switch_imgSVis(self):
        super().make_rgb_img(self.grid.imgs.get("visSeg"))
        self.repaint()

    def auto_seg(self, coefGrid=0):
        self.grid.cpuSeg(coefGrid=coefGrid)
        self.repaint()

    def fix_seg(self, width, length):
        self.grid.fixSeg(width=width, length=length)
        self.repaint()

    def align(self, method, axis=0):
        self.grid.agents.align(method=method, axis=axis)
        self.repaint()

    def distributed(self, axis=0, isEven=False):
        self.grid.agents.distributed(axis=axis, isEven=isEven)
        self.repaint()



# imports
import subprocess
import json
import sys
from urllib import request
from pkg_resources import parse_version

# basic imports
import sys
import os

# 3rd party imports
from PyQt6.QtWidgets import QApplication
from PyQt6.QtCore import QTimer
import qdarkstyle


# self imports
from grid.gridGUI import *
from grid.__init__ import __version__, __update__


def main():
    if "__main__" not in sys.argv[0]:
        # prevent from re-show welcome message in gridGUI
        # welcome message
        print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
        print("                 Welcome to GRID Ver.%s    " % __version__)
        print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
        print("Author      : James Chen <niche@vt.edu>    ")
        print("Last update : %s              " % __update__)
        print("User manual : https://poissonfish.github.io/GRID/")
        print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
        print("Recent update ")
        print("    - Now you can launch GRID by typing 'GRID' in terminal")
        print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")

    # self update
    try:
        url = "https://pypi.python.org/pypi/photo_grid/json"
        releases = json.loads(request.urlopen(url).read())["releases"]
        new_version = sorted(releases, key=parse_version, reverse=True)[0]
        if __version__ != new_version:
            # Dialog
            ans = None
            bol_ans = None
            possible_pos_ans = ["y", "Y", "yes"]
            possible_neg_ans = ["n", "N", "no"]

            while bol_ans is None:
                ans = input(
                    "A newer version of GRID (ver. %s) is now available, upgrade? (y/n) "
                    % new_version
                )
                if ans in possible_pos_ans:
                    bol_ans = True
                elif ans in possible_neg_ans:
                    bol_ans = False

            if bol_ans:
                subprocess.check_call(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        "photo_grid==%s" % new_version,
                        "--upgrade",
                    ]
                )
                print("\n")
                print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
                print("          Please re-launch GRID to finish the update")
                print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
                print("\n")
                quit()
    except Exception:
        print("\n")
        print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
        print("     Sorry, we currently have issue updating your GRID.")
        print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
        print("\n")

    app = QApplication(sys.argv)
    if "--light" not in sys.argv:
        app.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())

    grid = GRID_GUI()
    timer = QTimer()
    timer.timeout.connect(lambda: None)
    timer.start(100)
    app.exec()


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="photo_grid",
    version="1.3.10",
    description="A GUI for field segmentation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Poissonfish/GRID",
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows :: Windows 10",
    ],
    author="James Chen",
    author_email="niche@vt.edu",
    license="GPLv3",
    packages=["grid", "grid.gui"],
    include_package_data=True,
    install_requires=[
        "numpy",
        "pandas>=0.19.2",
        # data processing
        "h5py",
        "pyshp",
        # math, models
        "scikit-learn",
        "scipy",
        "matplotlib",
        # image processing
        "image",
        "opencv-python",
        "rasterio",
        # GUI
        "PyQt6",
        "qdarkstyle",
        # misc
        "tqdm",
    ],
    entry_points={"console_scripts": ["GRID=grid.__main__:main"]},
)



# Basic imports
import io
import os
import numpy as np
import pandas as pd
from urllib.request import urlopen
from PIL import Image

# 3-rd party imports
from PyQt6.QtCore import QFile, QIODevice
import h5py
import rasterio
import shapefile
import cv2

# self import
from grid.lib import (
    find_small_shape,
    initProgress,
    updateProgress,
    pltImShow,
    pltSegPlot,
    recover_scale,
)
from grid.dir import Dir


def loadImg(path):
    """
    ----------
    Parameters
    ----------
    path : str
           path to the image file

    -------
    Returns
    -------
    npImg : 3-d ndarray encoded in UINT8

    """

    # rasObj = rasterio.open(path)
    # nCh = rasObj.count
    # obj = initProgress(nCh, "loading channel 1")
    # if nCh < 3:
    #     npImg = np.zeros((rasObj.height, rasObj.width, 3), dtype="uint8")
    #     for i in range(3):
    #         npImg[:, :, i] = rasObj.read(1)
    #         updateProgress(obj, 1, "loading channel %d" %
    #                        (i+2 if i != (nCh-1) else i+1))
    # else:
    #     npImg = np.zeros((rasObj.height, rasObj.width, nCh), dtype="uint8")
    #     for i in range(nCh):
    #         npImg[:, :, i] = rasObj.read(i + 1)
    #         updateProgress(obj, "loading channel %d" % (i+2) if i != (nCh-1)
    # e
    # lse "Done")

    # return npImg
    rasObj = rasterio.open(path)
    ls_dtype = rasObj.dtypes
    nCh = rasObj.count
    prog = initProgress(nCh, name="Loading channel 1")
    if nCh < 3:
        npImg = np.zeros((rasObj.height, rasObj.width, 3), dtype="uint8")
        for i in range(3):
            npImg[:, :, i] = getBandUint8(rasObj.read(1), ls_dtype[1])
            if i != nCh - 1:
                updateProgress(prog, name="Loading channel %d" % (i + 2))
            else:
                updateProgress(prog, name="Done")
    else:
        npImg = np.zeros((rasObj.height, rasObj.width, nCh), dtype="uint8")
        for i in range(nCh):
            npImg[:, :, i] = getBandUint8(rasObj.read(i + 1), ls_dtype[i])
            if i != nCh - 1:
                updateProgress(prog, name="Loading channel %d" % (i + 2))
            else:
                updateProgress(prog, name="Done")

    # try to fetch geo info
    transform = rasObj.transform

    # fetch projection info
    try:
        crs = rasObj.crs.wkt
    except Exception:
        crs = False

    rasObj.close()

    # resize to avoid cv2 int16 limit (32767)
    h, w = find_small_shape(npImg.shape[:2])
    npImg = cv2.resize(npImg, (w, h))
    print("")
    print("GRID: Image was reshaped to (%d, %d)" % (h, w))

    # return
    return npImg, transform, crs


def getBandUint8(band, dtype):
    if "float" in dtype:
        band[band < 0] = 0
        band_int8 = (band - band.min()) * 255 / (np.quantile(band, 0.999) - band.min())
        band_int8[band_int8 > 255] = 255
        return band_int8
    else:
        return band


def loadImgWeb(URL):
    """
    ----------
    Parameters
    ----------
    URL : str
          URL to the UINT8-encoded image file

    -------
    Returns
    -------
    npImg : 3-d ndarray encoded in UINT8

    """

    with urlopen(URL) as url:
        file = io.BytesIO(url.read())
        npImg = np.array(Image.open(file), dtype="uint8")

    return npImg


def loadMap(path):
    """
    ----------
    Parameters
    ----------
    path : str
           path to the csv file

    -------
    Returns
    -------
    pdMap : Pandas dataframe or None if path is empty

    """

    try:
        pdMap = pd.read_csv(path, header=None)
    except Exception:
        pdMap = None

    return pdMap


def saveQImg(qimg, path):
    """
    ----------
    Parameters
    ----------
    qimg : qimage

    path : str
           path to the destination
    -------
    Returns
    -------
    None

    """

    qfile = QFile(path + ".jpg")
    qfile.open(QIODevice.WriteOnly)
    qimg.save(qfile, "JPG")
    qfile.close()


def saveDT(grid, path, prefix="GRID", simple=True):
    # save npy
    if not simple:
        np.save(os.path.join(path, prefix + "_image.npy"), grid.imgs.get("crop"))

    # get path
    pathDT = os.path.join(path, prefix + "_data.csv")

    # progress bar
    nD = grid.imgs.depth
    lsK = grid.imgs.paramKMs["lsSelect"]

    # grab info from GRID obj
    img = grid.imgs.get("crop").copy().astype(int)
    ch1Sub = 1 if img.shape[2] == 3 else 3  # replace NIR with Gr if it's RGB

    # intialize dataframe
    df = pd.DataFrame(columns=["var", "row", "col", "area_all", "area_veg"])

    # calculate index imgs
    dicIdx = dict(
        {
            "NDVI": (img[:, :, ch1Sub] - img[:, :, 0])
            / (img[:, :, ch1Sub] + img[:, :, 0] + 1e-8),
            "GNDVI": (img[:, :, ch1Sub] - img[:, :, 1])
            / (img[:, :, ch1Sub] + img[:, :, 1] + 1e-8),
            "CNDVI": (2 * img[:, :, ch1Sub] - img[:, :, 0] - img[:, :, 1])
            / (img[:, :, ch1Sub] + img[:, :, 0] + img[:, :, 1] + 1e-8),
            "RVI": img[:, :, ch1Sub] / (img[:, :, 0] + 1e-8),
            "GRVI": img[:, :, ch1Sub] / (img[:, :, 1] + 1e-8),
            "NDGI": (img[:, :, 1] - img[:, :, 0])
            / (img[:, :, 1] + img[:, :, 0] + 1e-8),
        }
    )

    # channel values
    for i in range(nD):
        name = "ch_%d" % (i + 1)
        dicIdx[name] = img[:, :, i]

    # # ratio of each cluster (k)
    # cluster = 0
    # for k in lsK:
    #     name = "cluster_%d" % cluster
    #     dicIdx[name] = (np.isin(grid.imgs.get("kmean"), i))*1
    #     cluster += 1

    # append columns based on the dict
    for key, _ in dicIdx.items():
        df[key] = None
        df[key + "_std"] = None

    # index data frame
    for row in range(grid.agents.nRow):
        for col in range(grid.agents.nCol):
            agent = grid.agents.get(row, col)
            if not agent or agent.isFake():
                continue
            try:
                entry = dict(var=str(agent.name), row=int(row + 1), col=int(col + 1))

                # compute valid ranges
                rg_row, rg_col = get_valid_range(agent, grid.imgs.get("bin"))

                # get selected pixels info
                imgBinAgent = grid.imgs.get("bin")[rg_row, :][:, rg_col]
                n_veg = imgBinAgent.sum()

                # load area info
                entry["area_all"] = len(rg_row) * len(rg_col)
                entry["area_veg"] = n_veg

                # append temp entry
                for key, imgIdx in dicIdx.items():
                    imgIdxAgent = imgIdx[rg_row, :][:, rg_col]
                    img_out = np.multiply(imgBinAgent, imgIdxAgent)
                    vec_out = img_out[img_out != 0].flatten()
                    entry[key] = vec_out.mean()
                    entry[key + "_std"] = vec_out.std()

                df.loc[len(df)] = entry

            except Exception as e:
                print("=====row: %d, col: %d=====" % (row, col))
                print(e)
                print("The plot is out of the borders")

    df = df[~df["var"].isnull()]

    # export
    df.to_csv(pathDT, index=False)


def savePlot(grid, path, prefix="GRID", simple=True):
    # raw-none
    pltSegPlot(
        grid.agents,
        grid.imgs.get("crop")[:, :, :3],
        path=path,
        prefix=prefix,
        filename="_raw.png",
    )
    if not simple:
        # raw-center
        pltSegPlot(
            grid.agents,
            grid.imgs.get("crop")[:, :, :3],
            isCenter=True,
            path=path,
            prefix=prefix,
            filename="_raw_centroid.png",
        )
        # raw-frame
        pltSegPlot(
            grid.agents,
            grid.imgs.get("crop")[:, :, :3],
            isRect=True,
            path=path,
            prefix=prefix,
            filename="_raw_border.png",
        )
        # raw-both
        pltSegPlot(
            grid.agents,
            grid.imgs.get("crop")[:, :, :3],
            isCenter=True,
            isRect=True,
            path=path,
            prefix=prefix,
            filename="_raw_both.png",
        )

    # cluster-none
    pltSegPlot(
        grid.agents,
        grid.imgs.get("kmean"),
        path=path,
        prefix=prefix,
        filename="_kmeans.png",
    )
    if not simple:
        # cluster-center
        pltSegPlot(
            grid.agents,
            grid.imgs.get("kmean"),
            isCenter=True,
            path=path,
            prefix=prefix,
            filename="_kmeans_centroid.png",
        )
        # cluster-frame
        pltSegPlot(
            grid.agents,
            grid.imgs.get("kmean"),
            isRect=True,
            path=path,
            prefix=prefix,
            filename="_kmeans_border.png",
        )
        # cluster-both
        pltSegPlot(
            grid.agents,
            grid.imgs.get("kmean"),
            isCenter=True,
            isRect=True,
            path=path,
            prefix=prefix,
            filename="_kmeans_both.png",
        )

    # binary-none
    pltSegPlot(
        grid.agents, grid.imgs.get("bin"), path=path, prefix=prefix, filename="_bin.png"
    )
    if not simple:
        # binary-center
        pltSegPlot(
            grid.agents,
            grid.imgs.get("bin"),
            isCenter=True,
            path=path,
            prefix=prefix,
            filename="_bin_centroid.png",
        )
        # binary-frame
        pltSegPlot(
            grid.agents,
            grid.imgs.get("bin"),
            isRect=True,
            path=path,
            prefix=prefix,
            filename="_bin_border.png",
        )
        # binary-both
        pltSegPlot(
            grid.agents,
            grid.imgs.get("bin"),
            isCenter=True,
            isRect=True,
            path=path,
            prefix=prefix,
            filename="_bin_both.png",
        )

    if not simple:
        # extraction-none
        pltSegPlot(
            grid.agents,
            grid.imgs.get("visSeg"),
            path=path,
            prefix=prefix,
            filename="_seg.png",
        )
        # extraction-center
        pltSegPlot(
            grid.agents,
            grid.imgs.get("visSeg"),
            isCenter=True,
            path=path,
            prefix=prefix,
            filename="_seg_centroid.png",
        )
        # extraction-frame
        pltSegPlot(
            grid.agents,
            grid.imgs.get("visSeg"),
            isCenter=True,
            isRect=True,
            path=path,
            prefix=prefix,
            filename="_seg_both.png",
        )

    # extraction-frame
    pltSegPlot(
        grid.agents,
        grid.imgs.get("visSeg"),
        isRect=True,
        path=path,
        prefix=prefix,
        filename="_seg_border.png",
    )
    # extraction-ID
    pltSegPlot(
        grid.agents,
        grid.imgs.get("visSeg"),
        isName=True,
        isRect=True,
        path=path,
        prefix=prefix,
        filename="_seg_ID.png",
    )


def saveH5(grid, path, prefix="GRID"):
    # get path
    pathH5 = os.path.join(path, prefix + ".h5")

    # create file
    with h5py.File(pathH5, "w"):
        None  # create file

    # index data frame
    img = grid.imgs.get("crop").copy()
    for row in range(grid.agents.nRow):
        for col in range(grid.agents.nCol):
            try:
                agent = grid.agents.get(row, col)
                if not agent or agent.isFake():
                    continue
                key = str(agent.name)

                # get ROI region
                rgY, rgX = get_valid_range(agent, grid.imgs.get("bin"))

                # compute kernel
                imgAll = img[:, rgX, :][rgY, :, :]
                imgBin = grid.imgs.get("bin")[:, rgX][rgY, :]
                imgFin = np.multiply(imgAll, np.expand_dims(imgBin, 2))
            except Exception as e:
                print("%s: The plot is out of the borders" % key)
                print(e)

            # export image
            try:
                with h5py.File(pathH5, "a") as f:
                    f.create_dataset(key, data=imgFin, compression="gzip")
                    f.create_dataset(key + "_raw", data=imgAll, compression="gzip")
            except Exception as e:
                print("Failed to save %s" % key)
                print(e)


def saveShape(grid, path, prefix="GRID"):
    pathDT = os.path.join(path, prefix + "_data.csv")
    pathSp = os.path.join(path, prefix)

    # info
    imgH = grid.map.imgH
    dt = pd.read_csv(pathDT)
    mat_H = grid.imgs.mat_H
    tiff_transform = grid.imgs.tiff_transform
    crs = grid.imgs.crs
    pts_crop = grid.imgs.pts_crop
    n_rot = grid.imgs.n_rot
    org = (int(grid.imgs.widthRs / 2), int(grid.imgs.heightRs / 2))

    with shapefile.Writer(pathSp) as f:
        # define fields
        cols = dt.columns

        for col in cols:
            instance = dt[col][0]

            try:
                float(instance)
                # integer, floating
                mode = "N"
                arg1, arg2 = 50, 50
            except ValueError:
                # characters
                mode = "C"
                arg1, arg2 = 50, 50

            if col == cols[0]:
                mode = "C"
                arg1, arg2 = 50, 50

            f.field(col, mode, arg1, arg2)

        for idx, entry in dt.iterrows():
            try:
                # get agents (the table is 1-index)
                row = entry["row"] - 1
                col = entry["col"] - 1
                agent = grid.agents.get(row, col)

                # polygon
                bn, bs, bw, be = get_valid_range(
                    agent, grid.imgs.get("bin"), is_border=True
                )
                pts_crop = [[bw, bn], [be, bn], [be, bs], [bw, bs]]

                # recover
                pts_rec = recover_scale(pts_crop, mat_H)

                # try remapping to Tiff coordinate
                pts_rec = [list(tiff_transform * pts_rec[i]) for i in range(4)]

                # input shape file
                f.poly([pts_rec])
            except Exception as e:
                print("%s: The plot is out of the borders" % entry["var"])
                print(e)

            # attributes
            dc = {c: entry[c] for c in dt.columns}
            f.record(**dict(dc))

    try:
        with open(pathSp + ".prj", "w") as f:
            f.write(crs)
    except Exception:
        print("No CRS found to write")


def get_valid_range(agent, img, is_border=False):
    # get border values
    bn = int(agent.getBorder(Dir.NORTH))
    bs = int(agent.getBorder(Dir.SOUTH))
    bw = int(agent.getBorder(Dir.WEST))
    be = int(agent.getBorder(Dir.EAST))

    # get image dimension
    imgH, imgW = img.shape

    # adjust to valid ranges
    if bn < 0:
        bn = 0
    if bw < 0:
        bw = 0
    if bs > imgH:
        # height
        bs = imgH
    if be >= imgW:
        # width
        be = imgW

    # make ranges
    rg_row = range(bn, bs)
    rg_col = range(bw, be)

    # return
    if is_border:
        return bn, bs, bw, be
    else:
        return rg_row, rg_col


# # create the PRJ file
# prj = open("%s.prj" % filename, "w")
# epsg = 'GEOGCS["WGS 84",'
# epsg += 'DATUM["WGS_1984",'
# epsg += 'SPHEROID["WGS 84",6378137,298.257223563]]'
# epsg += ',PRIMEM["Greenwich",0],'
# epsg += 'UNIT["degree",0.0174532925199433]]'
# prj.write(epsg)
# prj.close()
# WGS84-UTM


# PROJCS["WGS 84 / UTM zone 20N",
# GEOGCS["WGS 84",DATUM["WGS_1984",
# SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",-63],PARAMETER["scale_factor",0.9996],PARAMETER["false_easting",500000],PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","32620"]]'

# PROJCS["WGS 84 / UTM zone 20N",
# GEOGCS["WGS 84",
# DATUM["WGS_1984",
# SPHEROID["WGS 84", 6378137, 298.257223563,
# AUTHORITY["EPSG", "7030"]],
# AUTHORITY["EPSG", "6326"]],
# PRIMEM["Greenwich", 0,
#  AUTHORITY["EPSG", "8901"]],
# UNIT["degree", 0.0174532925199433,
# AUTHORITY["EPSG", "9122"]],
# AUTHORITY["EPSG", "4326"]],
#        PROJECTION["Transverse_Mercator"],
#        PARAMETER["latitude_of_origin", 0],
#        PARAMETER["central_meridian", -63],
#        PARAMETER["scale_factor", 0.9996],
#        PARAMETER["false_easting", 500000],
#        PARAMETER["false_northing", 0],
#        UNIT["metre", 1, AUTHORITY["EPSG", "9001"]],
#        AXIS["Easting", EAST], AXIS["Northing", NORTH], A
#        UTHORITY["EPSG", "32620"]]


# PROJCS["WGS_1984_UTM_Zone_20N",
#        GEOGCS["GCS_WGS_1984",
#               DATUM["D_WGS_1984",
#                     SPHEROID["WGS_1984", 6378137, 298.257223563]],
#               PRIMEM["Greenwich", 0],
#               UNIT["Degree", 0.017453292519943295]],
#        PROJECTION["Transverse_Mercator"],
#        PARAMETER["latitude_of_origin", 0],
#        PARAMETER["central_meridian", -63],
#        PARAMETER["scale_factor", 0.9996],
#        PARAMETER["false_easting", 500000],
#        PARAMETER["false_northing", 0], UNIT["Meter", 1]]



# basic imports
import numpy as np
import sys
import time
import math
import os
import pickle
import statistics

# 3rd party imports
from PyQt6.QtWidgets import *
from PyQt6.QtCore import *
from PyQt6.QtGui import *
from tqdm import tqdm, tqdm_gui
import cv2
from scipy.signal import convolve2d
from scipy.signal import find_peaks
import matplotlib.pyplot as plt
from matplotlib import patches
from matplotlib.lines import Line2D


def find_small_shape(shape_src, limits=32767, rate=0.95):
    h, w = shape_src
    while max(h, w) > limits:
        h *= rate
        w *= rate
    return int(h), int(w)


def doKMeans(img, k=3, features=[0]):
    """
    ----------
    Parameters
    ----------
    """

    # data type conversion for opencv
    ## select features
    img = img[:, :, features].copy()
    ## standardize
    img_max, img_min = img.max(axis=(0, 1)), img.min(axis=(0, 1)) - (1e-8)
    img = (img - img_min) / (img_max - img_min)
    ## convert to float32
    img_z = img.reshape((-1, img.shape[2])).astype(np.float32)

    # define criteria, number of clusters(K) and apply kmeans()
    criteria = (cv2.TERM_CRITERIA_MAX_ITER, 10, 1.0)
    param_k = dict(
        data=img_z,
        K=k,
        bestLabels=None,
        criteria=criteria,
        attempts=10,
        #    flags=cv2.KMEANS_RANDOM_CENTERS)
        flags=cv2.KMEANS_PP_CENTERS,
    )

    # KMEANS_RANDOM_CENTERS
    cv2.setRNGSeed(99163)
    _, img_k_temp, center = cv2.kmeans(**param_k)

    # Convert back
    img_k = img_k_temp.astype(np.uint8).reshape((img.shape[0], -1))

    # return
    return img_k, center


def blurImg(image, n, cutoff=0.5):
    image = smoothImg(image=image, n=n)
    return binarizeSmImg(image, cutoff=cutoff)


def smoothImg(image, n):
    """
    ----------
    Parameters
    ----------
    """

    kernel = np.array(([1, 4, 1], [4, 9, 4], [1, 4, 1]), dtype="int") / 29

    for _ in range(n):
        image = convolve2d(image, kernel, mode="same")

    return image


def binarizeSmImg(image, cutoff=0.5):
    """
    ----------
    Parameters
    ----------
    """
    imgOut = image.copy()
    imgOut[image > cutoff] = 1
    imgOut[image <= cutoff] = 0

    return imgOut.astype(int)


def find_angle(v1, v2):
    dot_product = np.dot(v1 / np.linalg.norm(v1), v2 / np.linalg.norm(v2))
    # cap dot_product
    if dot_product > 1:
        dot_product = 1
    elif dot_product < -1:
        dot_product = -1
    angle = np.arccos(dot_product) * 180 / np.pi
    if v1[0] * v2[1] - v1[1] * v2[0] > 0:
        return angle
    else:
        return 360 - angle


def sortPts(pts):
    """
    NOTE: qmap has inverse Y! UP has smaller y value
    """
    # get centroid to calculate vector
    pts_center = np.median(pts, axis=0)
    vec = pts - pts_center

    # get order
    vec_ag = [find_angle(vec[i], (-1, 0)) for i in range(4)]
    order = np.flip(np.argsort(vec_ag))
    # return sorted pts
    return np.array(pts)[order]


def cropImg(img, pts, resize=2048):
    """
    Perspectively project assigned area (pts) to a rectangle image
    -----
    param.
    -----
    img: 2-d numpy array
    pts: a vector of xy coordinate, length is 4. Must be in the order as:
         (NW, NE, SE, SW)
    """

    # define input coordinates
    pts = np.float32(pts)

    # assign sorted pts
    pt_NW, pt_NE, pt_SE, pt_SW = sortPts(pts)

    # estimate output dimension
    img_W = (euclidean(pt_NW, pt_NE) + euclidean(pt_SE, pt_SW)) / 2
    img_H = (euclidean(pt_SE, pt_NE) + euclidean(pt_SW, pt_NW)) / 2

    # resize output dimension
    shape = find_small_shape((img_W, img_H), limits=resize)

    # generate target point
    pts2 = np.float32(
        # NW,    NE,            SE,                   SW
        [[0, 0], [shape[0], 0], [shape[0], shape[1]], [0, shape[1]]]
    )

    # transformation
    H = cv2.getPerspectiveTransform(pts, pts2)
    dst = cv2.warpPerspective(img, H, (shape[0], shape[1]))
    dst = np.array(dst).astype(np.uint8)

    # return cropped image and H matrix
    return dst, H


def euclidean(p1, p2):
    """
    calculate Euclidean distance betweeen p1 and p2
    input can be either tuple or matrix
    """

    p1, p2 = np.array(p1), np.array(p2)
    return sum((p1 - p2) ** 2) ** (0.5)


def rotatePts(pts, angle, org=(0, 0)):
    """
    ----------
    Parameters
    ----------
    """
    ox, oy = org
    ptx = np.array([pts[i, 0] for i in range(len(pts))])
    pty = np.array([pts[i, 1] for i in range(len(pts))])
    qx = (
        ox
        + math.cos(math.radians(angle)) * (ptx - ox)
        - math.sin(math.radians(angle)) * (pty - oy)
    )
    qy = (
        oy
        + math.sin(math.radians(angle)) * (ptx - ox)
        + math.cos(math.radians(angle)) * (pty - oy)
    )
    qpts = [[qx[i], qy[i]] for i in range(len(pts))]
    return np.array(qpts)


# def rotate(origin, point, angle):
#     """
#     Rotate a point counterclockwise by a given angle around a given origin.

#     The angle should be given in radians.

#     @ contributed by Mark Dickinson
#     https://stackoverflow.com/users/270986/mark-dickinson
#     source:
#     https://stackoverflow.com/questions/34372480/rotate-point-about-another-point-in-degrees-python
#     """
#     ox, oy = origin
#     px, py = point

#     ag = np.pi / 180 * angle
#     qx = ox + np.cos(ag) * (px - ox) - np.sin(ag) * (py - oy)
#     qy = oy + np.sin(ag) * (px - ox) + np.cos(ag) * (py - oy)
#     return qx, qy

# === === === === === GRID pickle === === === === ===


def pickleGRID(obj, path):
    with open(path, "wb") as file:
        pickle.dump(obj, file, pickle.HIGHEST_PROTOCOL)


def getPickledGRID(path):
    with open(path, "rb") as file:
        obj = pickle.load(file)

    return obj


# === === === === === rank === === === === ===


def getRank(array):
    """
    ----------
    Results
    ----------
    [1,3,6,2,4] -> [4,2,0,3,1]
    """
    sort = np.array(array).argsort()
    rank = np.zeros(len(sort), dtype=int)
    rank[sort] = np.flip(np.arange(len(array)), axis=0)
    return rank


# === === === === === peak searching === === === === ===
def rotateNdArray(img, angle):
    depth = img.shape[2]
    list_img = []

    for i in range(depth):
        imgTemp = img[:, :, i]
        if imgTemp.max() > 1:
            val_max = 255
        else:
            val_max = 1

        # create border for the image
        imgTemp[:, 0:2] = val_max
        imgTemp[0:2, :] = val_max
        imgTemp[:, -2:] = val_max
        imgTemp[-2:, :] = val_max

        # padding
        sizePad = max(imgTemp.shape)
        imgP = np.pad(imgTemp, [sizePad, sizePad], "constant")

        # rotate
        pivot = (np.array(imgP.shape[:2]) / 2).astype(int).tolist()
        print("testse")
        matRot = cv2.getRotationMatrix2D(pivot, -angle, 1.0)
        imgR = cv2.warpAffine(
            imgP.astype(np.float32), matRot, imgP.shape, flags=cv2.INTER_LINEAR
        ).astype(np.uint8)

        # crop
        sigX = np.where(imgR.sum(axis=0) != 0)[0]
        sigY = np.where(imgR.sum(axis=1) != 0)[0]
        imgC = imgR[sigY[0] : sigY[-1], sigX[0] : sigX[-1]]

        # store output img
        list_img.append(imgC)

    imgFinal = np.array(list_img)
    imgFinal = np.swapaxes(imgFinal, 0, 1)
    imgFinal = np.swapaxes(imgFinal, 1, 2)

    return imgFinal


def rotateBinNdArray(img, angle):
    # create border for the image
    img[:, 0:2] = 1
    img[0:2, :] = 1
    img[:, -2:] = 1
    img[-2:, :] = 1

    # padding
    sizePad = max(img.shape)
    imgP = np.pad(img, [sizePad, sizePad], "constant").astype(np.float32)

    # rotate
    # check OpenCV version (4.5.2+ handle input format differently)
    ver_cv = cv2.__version__
    ver_cv_num = int(ver_cv.replace(".", "")[:3])
    is_old_cv = ver_cv_num < 452
    if is_old_cv:
        pivot = tuple((np.array(imgP.shape[:2]) / 2).astype(int))
    else:
        pivot = (np.array(imgP.shape[:2]) / 2).astype(int).tolist()
    # rotate images
    matRot = cv2.getRotationMatrix2D(pivot, -angle, 1.0)
    imgR = cv2.warpAffine(imgP, matRot, imgP.shape, flags=cv2.INTER_LINEAR).astype(
        np.uint8
    )

    # crop
    sigX = np.where(imgR.sum(axis=0) != 0)[0]
    sigY = np.where(imgR.sum(axis=1) != 0)[0]
    imgC = imgR[sigY[0] : sigY[-1], sigX[0] : sigX[-1]]

    # return
    return imgC


def rotateVec(vec, angle):
    deg = np.pi / 180
    x, y = vec[0], vec[1]
    xp = np.cos(deg * angle) * x - np.sin(deg * angle) * y
    yp = np.sin(deg * angle) * x + np.cos(deg * angle) * y
    return (xp, yp)


def invAffine(pt, affine, is3d=True):
    """
    convert GIS coordinate to (x, y)

    NOTE:
    # a = width of a pixel
    # b = row rotation(typically zero)
    # c = x-coordinate of the center of the upper-left pixel
    # d = column rotation(typically zero)
    # e = height of a pixel(typically negative)
    # f = y-coordinate of the center of the upper-left pixel
    """
    # transformation
    xg = (pt[0] - affine[2]) / affine[0]
    yg = (pt[1] - affine[5]) / affine[4]

    # return
    return (xg, yg, 1) if is3d else (xg, yg)


def recover_scale(mat_in, mat_H):
    """
    recover the cropped shaped into original scale

    parameters
    ----------
    mat_in: 4 x 2 matrix
    mat_H: 3 x 3 matrix
    """

    n_points = len(mat_in)
    # conver mat_in into 4 x 3 matrix
    mat_in = np.array([list(mat_in[i]) + [1] for i in range(4)])

    # solve recovered matrix
    mat_recover = np.matmul(np.linalg.inv(mat_H), mat_in.transpose())

    # transpose back to the right dimension (4 x 3)
    mat_recover = mat_recover.transpose()

    # extract the first 2 elements in each point (4 x 2)
    mat_recover = [mat_recover[i, :2] / mat_recover[i, 2] for i in range(n_points)]

    # return
    return np.array(np.matrix(mat_recover)).tolist()


def getFourierTransform(sig):
    sigf = abs(np.fft.fft(sig) / len(sig))
    return sigf[2 : int(len(sigf) / 2)]
    # return sigf[2:25]


def getCardIntercept(sig, angle, imgH=0):
    """
    transform signal to intercept
    """
    if angle == 0:
        return sig * 1
    else:
        coef = 1 / np.sin(np.pi / 180 * abs(angle))
        if angle < 0:
            return sig * coef
        else:
            return imgH - sig * coef


def getSigFromItc(itc, angle, imgH=0):
    """
    transform intercept to signal
    """
    if angle < 0 or angle > 90:
        sig = itc * np.sin(np.pi / 180 * abs(angle))
    elif angle > 0 and angle <= 90:
        sig = (imgH - itc) * np.sin(np.pi / 180 * abs(angle))
    else:
        # angle = 0
        sig = itc
    return sig


def getLineABC(slope, intercept):
    if np.isinf(slope):
        A = 1
        B = 0
        C = intercept
    else:
        A = slope
        B = -1
        C = -intercept
    return A, B, C


def solveLines(slope1, intercept1, slope2, intercept2):
    A1, B1, C1 = getLineABC(slope1, intercept1)
    A2, B2, C2 = getLineABC(slope2, intercept2)
    D = A1 * B2 - A2 * B1
    Dx = C1 * B2 - B1 * C2
    Dy = A1 * C2 - C1 * A2
    if D != 0:
        x, y = Dx / D, Dy / D
        return int(x), int(y)
    else:
        return False, False


def findPeaks(img, nPeaks=0, axis=1, nSmooth=100):
    """
    ----------
    Parameters
    ----------
    """

    # compute 1-D signal
    signal = img.mean(axis=(not axis) * 1)  # 0:nrow

    # ignore signals from iamge frame
    signal[:2] = [0, 0]
    signal[-2:] = [0, 0]

    # gaussian smooth
    for _ in range(int(len(signal) / 30)):
        signal = np.convolve(np.array([1, 2, 4, 2, 1]) / 10, signal, mode="same")

    # find primary peaks
    peaks, _ = find_peaks(signal)
    lsDiff = np.diff(peaks)
    medSig = statistics.median(signal[peaks])
    stdSig = np.array(signal[peaks]).std()
    medDiff = statistics.median(lsDiff)
    stdDiff = np.array(lsDiff).std()

    # get finalized peaks with distance constrain
    coef = 0.18 / (stdDiff / medDiff)  # empirical
    peaks, _ = find_peaks(signal, distance=medDiff - stdDiff * coef)
    # , prominence=(0.01, None))
    if nPeaks != 0:
        if len(peaks) > nPeaks:
            while len(peaks) > nPeaks:
                ls_diff = np.diff(peaks)
                idx_diff = np.argmin(ls_diff)
                idx_kick = (
                    idx_diff
                    if (signal[peaks[idx_diff]] < signal[peaks[idx_diff + 1]])
                    else (idx_diff + 1)
                )
                peaks = np.delete(peaks, idx_kick)
        elif len(peaks) < nPeaks:
            while len(peaks) < nPeaks:
                ls_diff = np.diff(peaks)
                idx_diff = np.argmax(ls_diff)
                peak_insert = (peaks[idx_diff] + peaks[idx_diff + 1]) / 2
                peaks = np.sort(np.append(peaks, int(peak_insert)))

    return peaks, signal


# === === === === Plotting === === === === ===


def qCross(x, y, painter, size=2):
    l1_st_x, l1_st_y = x - size, y - size
    l1_ed_x, l1_ed_y = x + size, y + size
    l2_st_x, l2_st_y = x - size, y + size
    l2_ed_x, l2_ed_y = x + size, y - size
    painter.drawLine(int(l1_st_x), int(l1_st_y), int(l1_ed_x), int(l1_ed_y))
    painter.drawLine(int(l2_st_x), int(l2_st_y), int(l2_ed_x), int(l2_ed_y))


def pltCross(x, y, size=3, width=1, color="red"):
    pt1X = [x - size, x + size]
    pt1Y = [y - size, y + size]
    line1 = Line2D(pt1X, pt1Y, linewidth=width, color=color)
    pt2X = [x - size, x + size]
    pt2Y = [y + size, y - size]
    line2 = Line2D(pt2X, pt2Y, linewidth=width, color=color)
    return line1, line2


def pltImShow(img, path=None, prefix="GRID", filename=".png"):
    if path is None:
        ax = plt.subplot(111)
        ax.imshow(img)
        plt.show()
    else:
        file = os.path.join(path, prefix + filename)
        if img.max() == 1:
            qimg = getBinQImg(img)
        elif img.max() < 100:
            qimg = getIdx8QImg(img, img.max() + 1)
        else:
            qimg = getRGBQImg(img)

        qimg.save(file, "PNG")


def pltSegPlot(
    agents,
    plotBase,
    isName=False,
    isRect=False,
    isCenter=False,
    path=None,
    prefix="GRID",
    filename=".png",
):
    if path is None:
        # CLI
        ax = plt.subplot(111)
        ax.imshow(plotBase)
        for row in range(agents.nRow):
            for col in range(agents.nCol):
                try:
                    agent = agents.get(row=row, col=col)
                    if not agent:
                        continue
                    recAg = agent.getQRect()
                    line1, line2 = pltCross(agent.x, agent.y, width=1)
                    ax.add_line(line1)
                    ax.add_line(line2)
                    if isRect:
                        rect = patches.Rectangle(
                            (recAg.x(), recAg.y()),
                            recAg.width(),
                            recAg.height(),
                            linewidth=1,
                            edgecolor="r",
                            facecolor="none",
                        )
                        ax.add_patch(rect)
                except Exception:
                    print("The plot is out of the borders")
        plt.show()
    else:
        # GUI
        file = os.path.join(path, prefix + filename)
        if plotBase.max() == 1:
            qimg = getBinQImg(plotBase)
        elif plotBase.max() < 10:
            qimg = getIdx8QImg(plotBase, plotBase.max() + 1)
        else:
            qimg = getRGBQImg(plotBase)

        pen = QPen()
        pen.setWidth(3)
        pen.setColor(Qt.red)
        painter = QPainter(qimg)
        painter.setPen(pen)
        painter.setBrush(Qt.transparent)
        for row in range(agents.nRow):
            for col in range(agents.nCol):
                try:
                    agent = agents.get(row, col)
                    center = agent.getCoordinate()
                    rect = agent.getQRect()
                    if isName:
                        text = "%s\n(%d, %d)" % (agent.name, row + 1, col + 1)
                        painter.drawText(rect, Qt.AlignmentFlag.AlignCenter, text)
                    if isRect:
                        painter.drawRect(rect)
                    if isCenter:
                        qCross(center[0], center[1], painter, size=3)
                except Exception:
                    print("The plot is out of the borders")
        painter.end()
        qimg.save(file, "PNG")


def pltImShowMulti(imgs, titles=None, vertical=False):
    nImgs = len(imgs)
    idxImg = 100 if vertical else 10
    idxLyt = 20 if vertical else 200

    plt.figure()
    for i in range(nImgs):
        idxPlot = idxImg * round(nImgs / 2) + idxLyt + (i + 1)
        plt.subplot(idxPlot)
        plt.imshow(imgs[i])
        try:
            plt.title(titles[i])
        except Exception:
            None

    plt.show()


def pltLinesPlot(gmap, agents, img):
    itcs = gmap.itcs
    slps = gmap.slps
    # plotting
    _, ax = plt.subplots()
    ax.imshow(img)
    for i in range(2):
        for intercept in itcs[i]:
            plotLine(ax, slps[i], intercept)
    for cAgents in agents:
        for agent in cAgents:
            line1, line2 = pltCross(agent.x, agent.y, width=2)
            ax.add_line(line1)
            ax.add_line(line2)

    plt.show()


def plotLine(axes, slope, intercept):
    if abs(slope) > 1e9:
        # vertical line
        y_vals = np.array(axes.get_ylim())
        x_vals = np.repeat(intercept, len(y_vals))
    else:
        # usual line
        x_vals = np.array(axes.get_xlim())
        y_vals = intercept + slope * x_vals
    axes.plot(x_vals, y_vals, "--", color="red")


def bugmsg(msg, title="DEBUG"):
    if "--test" in sys.argv:
        print("======%s=====" % title)
        print(msg)


# for GUI
def getRGBQImg(img):
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w * 3, QImage.Format_RGB888)
    return QPixmap(qImg)


def getBinQImg(img):
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w * 1, QImage.Format_Indexed8)
    qImg.setColor(0, qRgb(0, 0, 0))
    qImg.setColor(1, qRgb(241, 225, 29))
    return QPixmap(qImg)


def getIdx8QImg(img, k):
    colormap = [
        qRgb(228, 26, 28),
        qRgb(55, 126, 184),
        qRgb(77, 175, 74),
        qRgb(152, 78, 163),
        qRgb(255, 127, 0),
        qRgb(255, 255, 51),
        qRgb(166, 86, 40),
        qRgb(247, 129, 191),
        qRgb(153, 153, 153),
    ]
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w * 1, QImage.Format_Indexed8)
    for i in range(k):
        qImg.setColor(i, colormap[i])
    return QPixmap(qImg)


def getGrayQImg(img):
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w * 1, QImage.Format_Grayscale8)
    return QPixmap(qImg)


# progress bar
class GProg(QWidget):
    def __init__(self, size, name, widget):
        super().__init__()
        try:
            self._width = widget.width() / 5
            self._height = self._width / 16 * 5
            wgW, wgH = widget.width(), widget.height()
            self._pos = widget.pos()
        except Exception:
            self._width, self._height = 1, 1
            wgW, wgH = 1, 1
            self._pos = QPoint(1, 1)

        self.label = QLabel(name)
        font = QFont("Trebuchet MS", 20)
        self.label.setFont(font)
        self.bar = QProgressBar()
        self.bar.setRange(0, size)
        self.bar.setValue(0)
        self.layout = QVBoxLayout()
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.bar)
        self.setLayout(self.layout)
        self.move(
            int(self._pos.x() + (wgW - self._width) / 2),
            int(self._pos.y() + (wgH - self._height) / 2),
        )
        self.resize(int(self._width), int(self._height))
        self.show()
        self.repaint()
        QApplication.processEvents()

    def inc(self, n, name=None):
        self.bar.setValue(self.bar.value() + n)
        if name is not None:
            self.label.setText(name)
        self.repaint()
        QApplication.processEvents()

    def set(self, n, name=None):
        self.bar.setValue(n)
        if name is not None:
            self.label.setText(name)
        self.repaint()
        QApplication.processEvents()


def initProgress(size, name=None):
    if "__main__.py" in sys.argv[0]:
        # GUI
        widget = QApplication.activeWindow()
        obj = GProg(size, name, widget)
    else:
        # CLT
        obj = tqdm(total=size, postfix=name)

    return obj


def updateProgress(obj, n=1, name=None, flag=True):
    if (not flag) or (obj is None):
        return 0

    if "__main__.py" in sys.argv[0]:
        # GUI
        obj.inc(n, name)
    else:
        # CLT
        obj.set_postfix_str(name)
        obj.update(n)



# basic imports
import numpy as np
import pandas as pd

# self imports
from grid.lib import *
from grid.dir import Dir


class GAgent:
    def __init__(self):
        """
        ----------
        Parameters
        ----------
        """

        self.agents = []
        self.img = None
        self.shape = (0, 0)
        self.imgH, self.imgW = self.shape
        self.nRow, self.nCol = 0, 0
        self.rowFake, self.colFake = -1, -1
        self.coef = 0

        # progress bar
        self.flag = True
        self.subflag = True
        self.window = 1000

    def setup(self, gmap, gimg):
        """
        ----------
        Parameters
        ----------
        """

        self.agents = []
        self.img = gimg.get("binSeg")
        self.shape = self.img.shape
        self.imgH, self.imgW = self.shape

        # get nrow, ncol, and data.frame
        if not gimg.hasShp:
            self.nRow, self.nCol = gmap.nRow, gmap.nCol
            dt = gmap.dt
        else:
            """
            A case when a shapefile is loaded
            """
            # get transformation info
            affine = gimg.tiff_transform
            pts_crop = gimg.pts_crop
            mat_H = gimg.mat_H

            # load agents
            agents = gimg.f_shp.shapeRecords()

            # iteratively search for agents
            ls_id = []
            ls_row = []
            ls_col = []
            ls_ct = []
            ls_w, ls_n, ls_e, ls_s = [], [], [], []
            nrow = -1
            ncol = -1

            n_agents = len(agents)
            for i in range(n_agents):
                # extract agent info
                pts_ag = agents[i].shape.points
                att_ag = agents[i].record

                # find the remap pts
                pts_xy = [invAffine(pts_ag[i], affine) for i in range(4)]

                pts_crop_temp = np.matmul(
                    mat_H, np.float32(pts_xy).transpose()
                ).transpose()

                pts_crop = np.array(
                    [pts_crop_temp[i, :2] / pts_crop_temp[i, 2] for i in range(4)],
                    dtype=int,
                )
                # pts_crop = np.array(pts_crop_temp[:, :2], dtype=int)

                # find the centers
                pt_ct = np.mean(pts_crop, axis=0, dtype=int)
                ls_ct.append(pt_ct)
                # find the borders
                bd_w, bd_n = np.min(pts_crop, axis=0)
                bd_e, bd_s = np.max(pts_crop, axis=0)
                # examine the border
                bd_w = 0 if bd_w < 0 else bd_w
                bd_n = 0 if bd_n < 0 else bd_n
                bd_e = self.imgW - 1 if bd_e >= self.imgW else bd_e
                bd_s = self.imgH - 1 if bd_s >= self.imgH else bd_s
                # put in the list
                ls_w.append(bd_w)
                ls_n.append(bd_n)
                ls_e.append(bd_e)
                ls_s.append(bd_s)
                # get row and col
                row, col = np.array(att_ag[1:3], dtype=int)
                ls_row.append(row)
                ls_col.append(col)
                nrow = row if row > nrow else nrow
                ncol = col if col > ncol else ncol
                # get name
                name = att_ag[0]
                ls_id.append(name)

            # add by 1 as it's 0-index
            nrow += 1
            ncol += 1

            # assign
            self.nRow, self.nCol = nrow, ncol
            gmap.dt = pd.DataFrame(
                {
                    "var": ls_id,
                    "row": ls_row,
                    "col": ls_col,
                    "pt": ls_ct,
                    "bd_w": ls_w,
                    "bd_n": ls_n,
                    "bd_e": ls_e,
                    "bd_s": ls_s,
                }
            )
            dt = gmap.dt

        # initialize agents
        fr, fc = [], []
        for row in range(self.nRow):
            lsAgentsRow = []
            for col in range(self.nCol):
                try:
                    entry = dt[(dt.row == row) & (dt.col == col)].iloc[0]
                    ptX, ptY = entry["pt"]
                    name = entry["var"]
                except Exception:
                    # handle inconsistant # of rows/cols
                    fr.append(row)
                    fc.append(col)
                    ptX, ptY = (-1, -1)
                    name = "_FAKE"
                agent = Agent(name=name, row=row, col=col)
                self.setCoordinate(agent=agent, x=ptX, y=ptY)
                try:
                    # load boundaries if available
                    agent.setBorder(Dir.WEST, entry["bd_w"])
                    agent.setBorder(Dir.NORTH, entry["bd_n"])
                    agent.setBorder(Dir.EAST, entry["bd_e"])
                    agent.setBorder(Dir.SOUTH, entry["bd_s"])
                except Exception:
                    None
                lsAgentsRow.append(agent)
            self.agents.append(lsAgentsRow)

        self.rowFake, self.colFake = -1, -1
        # determine fake plots
        if len(fr) > 1 and np.array(fr).var() == 0:
            self.rowFake = fr[0]
        elif len(fc) > 1 and np.array(fc).var() == 0:
            self.colFake = fc[0]

    def get(self, row, col):
        """
        ----------
        Parameters
        ----------
        """

        if row < 0 or col < 0:
            # negative index
            return False
        else:
            try:
                return self.agents[int(row)][int(col)]
            except Exception:
                # outside frame
                return False

    def getNeib(self, row, col, dir):
        """
        ----------
        Parameters
        ----------
        """

        if dir == Dir.EAST:
            agent = self.get(row, col + 1)
            if not agent:
                # outside image
                return False
            elif col + 1 == self.colFake:
                # on fake line
                return agent
            elif agent.isFake():
                return self.get(row - 1, col + 1)
            else:
                return agent
        elif dir == Dir.SOUTH:
            agent = self.get(row + 1, col)
            if not agent:
                # outside image
                return False
            elif row + 1 == self.rowFake:
                # on fake line
                return agent
            elif agent.isFake():
                return self.get(row + 1, col - 1)
            else:
                return agent
        elif dir == Dir.WEST:
            agent = self.get(row, col - 1)
            if not agent:
                # outside image
                return False
            elif agent.isFake():
                return self.get(row - 1, col - 1)
            else:
                return agent
        elif dir == Dir.NORTH:
            agent = self.get(row - 1, col)
            if not agent:
                # outside image
                return False
            elif agent.isFake():
                return self.get(row + 1, col - 1)
            else:
                return agent

    def cpuPreDim(self, tol=5):
        """
        ----------
        Parameters
        ----------
        """
        self.isWork = True

        # prog = initProgress(self.nRow, name="Estimating dimentions")
        for row in range(self.nRow):
            # updateProgress(prog)
            for col in range(self.nCol):
                agentSelf = self.get(row, col)
                agentSelf.resetBorderHard()
                if not agentSelf or agentSelf.isFake():
                    continue
                rgTemp = dict()
                for axis in [0, 1]:
                    # extract direction info and 1dImg
                    dir1 = Dir(axis)  # axis:0, return N(0) and S(2)
                    dir2 = Dir(axis + 2)  # axis:1, return W(1) and E(3)
                    axisRev = (not axis) * 1
                    # img1d = self.img[agentSelf.y,
                    #                  :] if axis else self.img[:, agentSelf.x]
                    # extract agents info
                    ptSelf = agentSelf.getCoordinate()[axisRev]
                    agentNeig1 = self.getNeib(row, col, dir1)
                    agentNeig2 = self.getNeib(row, col, dir2)
                    # if both neighbors exists
                    if (
                        (agentNeig1 != 0)
                        and (agentNeig2 != 0)
                        and (not agentNeig1.isFake())
                        and (not agentNeig2.isFake())
                    ):
                        ptNeig1 = agentNeig1.getCoordinate()[axisRev]
                        ptNeig2 = agentNeig2.getCoordinate()[axisRev]
                        ptMid = int((ptNeig1 + ptNeig2) / 2)
                        ptBd1 = int((ptNeig1 + ptMid) / 2)
                        ptBd2 = int((ptNeig2 + ptMid) / 2)
                    # if only left/up side exist
                    elif agentNeig1:
                        ptNeig1 = agentNeig1.getCoordinate()[axisRev]
                        ptBd1 = int((ptSelf + ptNeig1) / 2)
                        ptBd2 = self.img.shape[axis]
                        self.setBorder(agentSelf, dir2, ptBd2)
                    # if only right/down side exist
                    elif agentNeig2:
                        ptNeig2 = agentNeig2.getCoordinate()[axisRev]
                        ptBd1 = 0
                        ptBd2 = int((ptSelf + ptNeig2) / 2)
                        self.setBorder(agentSelf, dir1, ptBd1)
                    # if neither side exist
                    elif self.nRow == 1 or self.nCol == 1:
                        ptBd1 = 0
                        ptBd2 = self.img.shape[axis]
                        self.setBorder(agentSelf, dir1, ptBd1)
                        self.setBorder(agentSelf, dir2, ptBd2)

                    # check if invade to its neigbor
                    # new_centroid = (ptBd1 + ptBd2) / 2
                    # if new_centroid >=

                    # # get predim
                    # # negative side (neighber 1)
                    # pt_cur = ptSelf
                    # tol_cur = 0
                    # while (tol_cur < tol) & (pt_cur > ptBd1):
                    #     try:
                    #         img_val = img1d[pt_cur]
                    #     except Exception:
                    #         break
                    #     tol_cur += 1 if img_val == 0 else - tol_cur  # else reset to 0
                    #     pt_cur -= 1
                    # rgTemp[dir1.name] = pt_cur
                    # # positive side (neighber 2)
                    # pt_cur = ptSelf
                    # tol_cur = 0
                    # while (tol_cur < tol) & (ptBd2 > pt_cur):
                    #     try:
                    #         img_val = img1d[pt_cur]
                    #     except Exception:
                    #         break
                    #     tol_cur += 1 if img_val == 0 else - tol_cur  # else reset to 0
                    #     pt_cur += 1
                    # rgTemp[dir2.name] = pt_cur

                    rgTemp[dir1.name] = ptBd1
                    rgTemp[dir2.name] = ptBd2

                agentSelf.setPreDim(rgTemp)

    def autoSeg(self, coefGrid=0.2):
        """
        ----------
        Parameters
        ----------
        """
        self.coef = coefGrid

        # reset the border first
        self.resetBorder()

        # progress bar
        prog = None
        if self.flag:
            self.flag = False
            prog = initProgress(self.nRow, name="Segmenting")

        # loop over rows and cols
        for row in range(self.nRow):
            updateProgress(prog, flag=self.subflag)
            for col in range(self.nCol):
                agentSelf = self.get(row, col)
                for dir in list([Dir.EAST, Dir.SOUTH]):
                    agentNeib = self.getNeib(row, col, dir)
                    dirNeib = list(Dir)[(dir.value + 2) % 4]  # reverse the direction
                    if (
                        not agentSelf
                        or agentSelf.isFake()
                        or not agentNeib
                        or agentNeib.isFake()
                    ):
                        continue
                    # reset agent border
                    agentSelf.border[dir.name] = agentSelf.border_reset[dir.name]
                    agentNeib.border[dirNeib.name] = agentNeib.border_reset[
                        dirNeib.name
                    ]
                    # calculate border
                    if dir == Dir.EAST:
                        dist_agents = abs(agentSelf.x - agentNeib.x)
                    else:
                        dist_agents = abs(agentSelf.y - agentNeib.y)

                    while (agentNeib.getBorder(dirNeib) - agentSelf.getBorder(dir)) > 1:
                        scASelf = agentSelf.getScoreArea(dir, self.img)
                        scGSelf = agentSelf.getScoreGrid(dir) / dist_agents
                        scANeib = agentNeib.getScoreArea(dirNeib, self.img)
                        scGNeib = agentNeib.getScoreGrid(dirNeib) / dist_agents
                        # except Exception as e:
                        scoreSelf = scASelf - (scGSelf * coefGrid)
                        scoreNeib = scANeib - (scGNeib * coefGrid)
                        if scoreSelf > scoreNeib:
                            self.updateBorder(agentSelf, dir, 1)
                        elif scoreSelf < scoreNeib:
                            self.updateBorder(agentNeib, dirNeib, -1)
                        else:
                            self.updateBorder(agentSelf, dir, 1)
                            self.updateBorder(agentNeib, dirNeib, -1)

        if self.subflag:
            self.subflag = False
            QTimer.singleShot(self.window, lambda: setattr(self, "flag", True))
            QTimer.singleShot(self.window, lambda: setattr(self, "subflag", True))

        # rescue plots on fake line
        if self.rowFake != -1:
            for col in range(1, self.nCol):
                agent = self.get(self.rowFake, col)
                if not agent.isFake():
                    agentNeig = self.get(self.rowFake - 1, col - 1)
                    agent.setBorder(Dir.WEST, agentNeig.getBorder(Dir.EAST))
        elif self.colFake != -1:
            for row in range(1, self.nRow):
                agent = self.get(row, self.colFake)
                if not agent.isFake():
                    agentNeig = self.get(row - 1, self.colFake - 1)
                    agent.setBorder(Dir.NORTH, agentNeig.getBorder(Dir.SOUTH))

    def fixSeg(self, width, length):
        w_unit = (self.imgW / self.nCol) / 100
        l_unit = (self.imgH / self.nRow) / 100
        w_side = round(width / 2 * w_unit)
        l_side = round(length / 2 * l_unit)
        for row in range(self.nRow):
            for col in range(self.nCol):
                agent = self.get(row, col)
                if not agent or agent.isFake():
                    continue
                agent.resetBorder()
                # set border
                if (row == (self.nRow - 1)) & (length == 100):
                    self.setBorder(agent, Dir.SOUTH, self.imgH)
                else:
                    self.setBorder(agent, Dir.SOUTH, agent.y + l_side)
                if (row == 0) & (length == 100):
                    self.setBorder(agent, Dir.NORTH, 0)
                else:
                    self.setBorder(agent, Dir.NORTH, agent.y - l_side)
                if (col == (self.nCol - 1)) & (width == 100):
                    self.setBorder(agent, Dir.EAST, self.imgW)
                else:
                    self.setBorder(agent, Dir.EAST, agent.x + w_side)
                if (col == 0) & (width == 100):
                    self.setBorder(agent, Dir.WEST, 0)
                else:
                    self.setBorder(agent, Dir.WEST, agent.x - w_side)

    def setCoordinate(self, agent, x, y):
        agent.setCoordinate(x, y)
        self.checkBorder(agent)

    def updateCoordinate(self, agent, value, axis):
        agent.updateCoordinate(value, axis)
        self.checkBorder(agent)

    def setBorder(self, agent, dir, value):
        agent.setBorder(dir, value)
        self.checkBorder(agent)

    def updateBorder(self, agent, dir, value):
        agent.updateBorder(dir, value)
        self.checkBorder(agent)

    def resetCoordinate(self):
        for row in range(self.nRow):
            for col in range(self.nCol):
                try:
                    self.get(row=row, col=col).resetCoordinate()
                except Exception:
                    None

    def resetBorder(self):
        for row in range(self.nRow):
            for col in range(self.nCol):
                try:
                    self.get(row=row, col=col).resetBorder()
                except Exception:
                    None

    def checkBorder(self, agent):
        if agent.border[Dir.NORTH.name] < 0:
            agent.border[Dir.NORTH.name] = 0
        if agent.border[Dir.WEST.name] < 0:
            agent.border[Dir.WEST.name] = 0
        if agent.border[Dir.SOUTH.name] >= self.imgH:
            agent.border[Dir.SOUTH.name] = self.imgH - 1
        if agent.border[Dir.EAST.name] >= self.imgW:
            agent.border[Dir.EAST.name] = self.imgW - 1

    def align(self, method, axis=0):
        """ """
        if method == 0:
            # None
            for row in range(self.nRow):
                for col in range(self.nCol):
                    agent = self.get(row=row, col=col)
                    if not agent or agent.isFake():
                        continue
                    if axis == 0:
                        dist = agent.y_reset - agent.y
                    elif axis == 1:
                        dist = agent.x_reset - agent.x
                    self.updateCoordinate(agent, dist, axis=axis)
        elif method == 1:
            # Top/Left
            if axis == 0:
                for row in range(self.nRow):
                    val = 1e10
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val_temp = agent.y
                        val = val_temp if val_temp < val else val
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.y
                        self.updateCoordinate(agent, dist, axis=axis)
            elif axis == 1:
                for col in range(self.nCol):
                    val = 1e10
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val_temp = agent.x
                        val = val_temp if val_temp < val else val
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.x
                        self.updateCoordinate(agent, dist, axis=axis)
        elif method == 3:
            # Bottom/Right
            if axis == 0:
                for row in range(self.nRow):
                    val = -1
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val_temp = agent.y
                        val = val_temp if val_temp > val else val
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.y
                        self.updateCoordinate(agent, dist, axis=axis)
            elif axis == 1:
                for col in range(self.nCol):
                    val = -1
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val_temp = agent.x
                        val = val_temp if val_temp > val else val
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.x
                        self.updateCoordinate(agent, dist, axis=axis)
        elif method == 2:
            # Middle/Center
            if axis == 0:
                for row in range(self.nRow):
                    val = 0
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val += agent.y
                    val = int(val / (self.nCol))
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.y
                        self.updateCoordinate(agent, dist, axis=axis)
            elif axis == 1:
                for col in range(self.nCol):
                    val = 0
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val += agent.x
                    val = int(val / (self.nRow))
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.x
                        self.updateCoordinate(agent, dist, axis=axis)

    def pan(self, axis, target, value):
        if axis == 0:
            for col in range(self.nCol):
                agent = self.get(row=target, col=col)
                # dist = value - agent.y
                self.updateCoordinate(agent, value, axis=axis)
        elif axis == 1:
            for row in range(self.nRow):
                agent = self.get(row=row, col=target)
                # dist = value - agent.x
                self.updateCoordinate(agent, value, axis=axis)

    def distributed(self, axis, isEven):
        if isEven:
            if axis == 0:
                # row
                y_North = self.get(row=0, col=0).y
                y_South = self.get(row=self.nRow - 1, col=0).y
                dist = y_South - y_North
                pos_new = np.arange(y_North, y_South, dist / (self.nRow - 1))
                pos_new = np.append(pos_new, y_South)
                for row in range(self.nRow):
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = pos_new[row] - agent.y
                        self.updateCoordinate(agent=agent, value=dist, axis=0)
            else:
                # col
                x_West = self.get(row=0, col=0).x
                x_East = self.get(row=0, col=self.nCol - 1).x
                dist = x_East - x_West
                pos_new = np.arange(x_West, x_East, dist / (self.nCol - 1))
                pos_new = np.append(pos_new, x_East)
                for col in range(self.nCol):
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = pos_new[col] - agent.x
                        self.updateCoordinate(agent=agent, value=dist, axis=1)
        else:
            for row in range(self.nRow):
                for col in range(self.nCol):
                    agent = self.get(row=row, col=col)
                    if not agent or agent.isFake():
                        continue
                    if axis == 0:
                        dist = agent.y_reset - agent.y
                    elif axis == 1:
                        dist = agent.x_reset - agent.x
                    self.updateCoordinate(agent=agent, value=dist, axis=axis)


class Agent:
    def __init__(self, name, row, col):
        """ """
        self.name = name
        self.row, self.col = row, col
        self.y, self.x = 0, 0
        self.y_float, self.x_float = 0.0, 0.0  # range from 0 to 0.9999
        self.y_reset, self.x_reset = 0, 0
        self.pre_rg_W, self.pre_rg_H = range(0), range(0)
        self.border, self.border_reset = dict(), dict()
        for dir in list([Dir.NORTH, Dir.EAST, Dir.SOUTH, Dir.WEST]):
            self.border[dir.name] = 0
            self.border_reset[dir.name] = 0

    def getCoordinate(self):
        """ """
        return self.x, self.y

    def getPreDim(self, isHeight=True):
        """ """
        return self.pre_rg_H if isHeight else self.pre_rg_W

    def getBorder(self, dir):
        return self.border[dir.name]

    def getScoreArea(self, dir, img):
        """
        Will ragne from 0 to 1
        """
        isH = dir.value % 2  # E->1, S->0
        rg = self.getPreDim(isHeight=isH)
        bd = self.getBorder(dir)
        return img[rg, bd].mean() if isH else img[bd, rg].mean()

    def getScoreGrid(self, dir):
        """
        Will ragne from 0 to 1
        """
        isWE = dir.value % 2  # is W, E or N, S
        pt_center = self.x if isWE else self.y
        pt_cur = self.getBorder(dir)
        return abs(pt_cur - pt_center)

    def setCoordinate(self, x, y):
        """ """
        self.x, self.y = int(x), int(y)
        self.x_reset, self.y_reset = int(x), int(y)
        self.setBorder(Dir.NORTH, y)
        self.setBorder(Dir.SOUTH, y)
        self.setBorder(Dir.WEST, x)
        self.setBorder(Dir.EAST, x)

    def setPreDim(self, rg):
        """ """
        self.pre_rg_W = range(rg["WEST"], rg["EAST"])
        self.pre_rg_H = range(rg["NORTH"], rg["SOUTH"])
        # self.x = int((rg['EAST']+rg['WEST'])/2)
        # self.y = int((rg['NORTH']+rg['SOUTH'])/2)
        # self.x_reset, self.y_reset = self.x, self.y
        for dir in list([Dir.NORTH, Dir.WEST, Dir.SOUTH, Dir.EAST]):
            self.border_reset[dir.name] = self.border[dir.name]

    def setBorder(self, dir, value):
        """ """
        self.border[dir.name] = int(value)

    def updateBorder(self, dir, value):
        """ """
        self.border[dir.name] += int(value)

    def updateCoordinate(self, value, axis=0):
        """
        decompose value to integer and floating parts
        """
        if axis == 0:
            self.y_float += value
            dy, self.y_float = self.y_float // 1, self.y_float % 1
            self.y += dy
            self.border[Dir.NORTH.name] += dy
            self.border[Dir.SOUTH.name] += dy
        elif axis == 1:
            self.x_float += value
            dx, self.x_float = self.x_float // 1, self.x_float % 1
            self.x += dx
            self.border[Dir.WEST.name] += dx
            self.border[Dir.EAST.name] += dx

    def getQRect(self):
        """
        GUI SPECIFIC
        """
        x = self.getBorder(Dir.WEST)
        y = self.getBorder(Dir.NORTH)
        w = self.getBorder(Dir.EAST) - x
        h = self.getBorder(Dir.SOUTH) - y
        return QRect(int(x), int(y), int(w), int(h))

    def resetCoordinate(self):
        self.x = self.x_reset
        self.y = self.y_reset
        self.resetBorder()

    def resetBorder(self):
        for dir in list([Dir.NORTH, Dir.WEST, Dir.SOUTH, Dir.EAST]):
            self.border[dir.name] = self.border_reset[dir.name]

    def resetBorderHard(self):
        self.setBorder(Dir.WEST, self.x)
        self.setBorder(Dir.EAST, self.x)
        self.setBorder(Dir.NORTH, self.y)
        self.setBorder(Dir.SOUTH, self.y)

    def isFake(self):
        return self.name == "_FAKE"
